@@ -105,6 +105,7 @@ my @valid_severities=qw(wishlist minor normal important
 			serious grave critical);
 
 my $browser;  # Will set if necessary
+my $btsserver='bugs.debian.org';
 my $btsurl='http://bugs.debian.org/';
 my $btscgiurl='http://bugs.debian.org/cgi-bin/';
 my $btscgipkgurl='http://bugs.debian.org/cgi-bin/pkgreport.cgi';
@@ -140,14 +141,15 @@ In general, the command line interface is the same as what you would write
 in a mail to control@bugs.debian.org, just prefixed with "bts". For
 example:
 
- % bts close 85942 1.1
  % bts severity 69042 normal
  % bts merge 69042 43233
  % bts retitle 69042 blah blah
 
 A few additional commands have been added for your convenience, and this
 program is less strict about what constitutes a valid bug number. For example,
-"close Bug#85942" is understood, as is "close #85942".
+"severity Bug#85942 normal" is understood, as is "severity #85942 normal".
+(Of course, your shell may regard "#" as a comment character though, so you
+may need to quote it!)
 
 Also, for your convenience, this program allows you to abbreviate commands
 to the shortest unique substring (similar to how cvs lets you abbreviate
@@ -158,8 +160,8 @@ your shell does not strip out the comment in a command like
 "bts severity 30321 normal #inflated severity", then this program is smart
 enough to figure out where the comment is, and include it in the email.
 Note that most shells do strip out such comments before they get to the
-program, unless the comment is quoted.  (Note that something like "bts
-close #85942" will not be treated as a comment!)
+program, unless the comment is quoted.  (Something like "bts
+severity #85942 normal" will not be treated as a comment!)
 
 You can specify multiple commands by separating them with a single dot,
 rather like B<update-rc.d>; a single comma may also be used; all the
@@ -225,6 +227,11 @@ will be replaced by the name of the mbox file.  The command will be split
 on white space and will not be passed to a shell.  Default is 'mutt -f %s'.
 (Also, %% will be substituted by a single % if this is needed.)
 
+=item --sendmail=SENDMAILCMD
+
+Specify the sendmail command.  The command will be split on white
+space and will not be passed to a shell.  Default is '/usr/sbin/sendmail'.
+
 =item -f, --force-refresh
 
 Download a bug report again, even if it does not appear to have
@@ -259,6 +266,7 @@ my $caching=1;
 my $cachemode='min';
 my $refreshmode=0;
 my $mailreader='mutt -f %s';
+my $sendmailcmd='/usr/sbin/sendmail';
 
 # Next, read read configuration files and then command line
 # The next stuff is boilerplate
@@ -274,6 +282,7 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
 		       'BTS_CACHE_MODE' => 'min',
 		       'BTS_FORCE_REFRESH' => 'no',
 		       'BTS_MAIL_READER' => 'mutt -f %s',
+		       'BTS_SENDMAIL_COMMAND' => '/usr/sbin/sendmail',
 		       );
     my %config_default = %config_vars;
     
@@ -300,6 +309,19 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
 	or $config_vars{'BTS_FORCE_REFRESH'}='no';
     $config_vars{'BTS_MAIL_READER'} =~ /\%s/
 	or $config_vars{'BTS_MAIL_READER'}='mutt -f %s';
+    $config_vars{'BTS_SENDMAIL_COMMAND'} =~ /./
+	or $config_vars{'BTS_SENDMAIL_COMMAND'}='/usr/sbin/sendmail';
+
+    if ($config_vars{'BTS_SENDMAIL_COMMAND'} ne '/usr/sbin/sendmail') {
+	my $cmd = (split ' ', $config_vars{'BTS_SENDMAIL_COMMAND'})[0];
+	unless ($cmd =~ /^[A-Za-z0-9_\-\+\.\/]*$/) {
+	    warn "BTS_SENDMAIL_COMMAND contained funny characters: $cmd\nReverting to default value /usr/sbin/sendmail\n";
+	    $config_vars{'BTS_SENDMAIL_COMMAND'}='/usr/sbin/sendmail';
+	} elsif (system("command -v $cmd >/dev/null 2>&1") != 0) {
+	    warn "BTS_SENDMAIL_COMMAND $cmd could not be executed.\nReverting to default value /usr/sbin/sendmail\n";
+	    $config_vars{'BTS_SENDMAIL_COMMAND'}='/usr/sbin/sendmail';
+	}
+    }
 
     foreach my $var (sort keys %config_vars) {
 	if ($config_vars{$var} ne $config_default{$var}) {
@@ -314,6 +336,7 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
     $cachemode = $config_vars{'BTS_CACHE_MODE'};
     $refreshmode = $config_vars{'BTS_FORCE_REFRESH'} eq 'yes' ? 1 : 0;
     $mailreader = $config_vars{'BTS_MAIL_READER'};
+    $sendmailcmd = $config_vars{'BTS_SENDMAIL_COMMAND'};
 }
 
 if (exists $ENV{'BUGSOFFLINE'}) {
@@ -321,7 +344,7 @@ if (exists $ENV{'BUGSOFFLINE'}) {
 }
 
 my ($opt_help, $opt_version, $opt_noconf);
-my ($opt_cachemode, $opt_mailreader);
+my ($opt_cachemode, $opt_mailreader, $opt_sendmail);
 my $opt_cachedelay=5;
 my $mboxmode = 0;
 my $quiet=0;
@@ -337,6 +360,7 @@ GetOptions("help|h" => \$opt_help,
 	   "cache-delay=i" => \$opt_cachedelay,
 	   "m|mbox" => \$mboxmode,
 	   "mailreader|mail-reader=s" => \$opt_mailreader,
+	   "sendmail=s" => \$opt_sendmail,
 	   "f" => \$refreshmode,
 	   "force-refresh!" => \$refreshmode,
 	   "q|quiet+" => \$quiet,
@@ -358,6 +382,21 @@ if ($opt_mailreader) {
     }
 }
 
+if ($opt_sendmail) {
+    if ($opt_sendmail ne '/usr/sbin/sendmail'
+	and $opt_sendmail ne $sendmailcmd) {
+	my $cmd = (split ' ', $opt_sendmail)[0];
+	unless ($cmd =~ /^[A-Za-z0-9_\-\+\.\/]*$/) {
+	    warn "--sendmail command contained funny characters: $cmd\nReverting to default value $sendmailcmd\n";
+	    unset $opt_sendmail;
+	} elsif (system("command -v $cmd >/dev/null 2>&1") != 0) {
+	    warn "--sendmail command $cmd could not be executed.\nReverting to default value $sendmailcmd\n";
+	    unset $opt_sendmail;
+	}
+    }
+}
+$sendmailcmd = $opt_sendmail if $opt_sendmail;
+
 if ($opt_cachemode) {
     if ($opt_cachemode =~ /^(min|mbox|full)$/) {
 	$cachemode=$opt_cachemode;
@@ -427,13 +466,17 @@ exit 0;
 =head1 COMMANDS
 
 For full details about the commands, see the BTS documentation.
-L<http://bugs.debian.org/server-control>
+L<http://www.debian.org/Bugs/server-control>
 
 =over 4
 
 =item show [options] [<bug number> | <package> | <maintainer> | : ] [opt=val ...]
 
-=item show [options] [src:<package> | from:<submitter> | tag:<tag> ] [opt=val ...]
+=item show [options] [src:<package> | from:<submitter>] [opt=val ...]
+
+=item show [options] [tag:<tag> | usertag:<tag> ] [opt=val ...]
+
+=item show [release-critical | release-critical/... | RC]
 
 This is a synonym for bts bugs.
 
@@ -445,7 +488,11 @@ sub bts_show {
 
 =item bugs [options] [<bug number> | <package> | <maintainer> | : ] [opt=val ..]
 
-=item bugs [options] [src:<package> | from:<submitter> | tag:<tag> ] [opt=val ..]
+=item bugs [options] [src:<package> | from:<submitter>] [opt=val ..]
+
+=item bugs [options] [tag:<tag> | usertag:<tag> ] [opt=val ..]
+
+=item bugs [release-critical | release-critical/... | RC]
 
 Display the page listing the requested bugs in a web browser using
 L<sensible-browser(1)>.
@@ -491,6 +538,12 @@ Display the bugs for the submitter email address <submitter>.
 
 Display the bugs which are tagged with <tag>.
 
+=item usertag:<tag>
+
+Display the bugs which are tagged with usertag <tag>.  See the BTS
+documentation for more information on usertags.  This will require the
+use of a user=<email> option.
+
 =item :
 
 Details of the bug tracking system itself, along with a bug-request
@@ -498,6 +551,13 @@ page with more options than this script, can be found on
 http://bugs.debian.org/.  This page itself will be opened if the
 command 'bts bugs :' is used.
 
+=item release-critical, RC
+
+Display the front page of the release-critical pages on the BTS.  This
+is a synonym for http://bugs.debian.org/release-critical/index.html.
+It is also possible to say release-critical/debian/main.html and the like.
+RC is a synonym for release-critical/other/all.html.
+
 =back
 
 After the argument specifying what to display, you can optionally
@@ -546,7 +606,7 @@ BROWSER='mozilla -raise -remote "openURL(%s,new-window)":links'
 =cut
 
 sub bts_bugs {
-    @ARGV = @_;
+    @ARGV = @_; # needed for GetOptions
     my ($sub_offlinemode, $sub_caching, $sub_mboxmode, $sub_mailreader);
     GetOptions("o" => \$sub_offlinemode,
 	       "offline!" => \$sub_offlinemode,
@@ -576,7 +636,7 @@ sub bts_bugs {
 	}
     }
 
-    my $url = shift;
+    my $url = sanitizething(shift);
     if (! $url) {
 	if (defined $ENV{'DEBEMAIL'}) {
 	    $url=$ENV{'DEBEMAIL'};
@@ -588,14 +648,18 @@ sub bts_bugs {
 	    }
 	}
     }
-    if ($url =~ /^.*\s+<(.*)>$/) { $url = $1; }
+    if ($url =~ /^.*\s<(.*)>\s*$/) { $url = $1; }
     $url =~ s/^:$//;
-    if (@_) {
-	$url = join(";", $url, @_);
-	$url =~ s/:/=/g;
-	$url =~ s/;tag=/;include=/;
+
+    # Are there any options?
+    my $urlopts = '';
+    if (@_) { 
+	$urlopts = join(";", '', @_); # so it'll be ";opt1=val1;opt2=val2"
+	$urlopts =~ s/:/=/g;
+	$urlopts =~ s/;tag=/;include=/;
     }
-    browse($url);
+    
+    browse($url, $urlopts);
 
     # revert options
     if (defined $sub_offlinemode) {
@@ -693,7 +757,7 @@ Change the submitter address of a bug or a number of bugs, with `!' meaning
 sub bts_submitter {
     @_ or die "bts submitter: change submitter of what bug?\n";
     my $submitter=pop;
-    if ($submitter !~ /\@/) {
+    if ($submitter !~ /\@/ and $submitter ne '!') {
 	die "bts submitter: change submitter to what?\n";
     }
     foreach (@_) {
@@ -730,7 +794,7 @@ sub bts_reassign {
     }
 }
 
-=item found <bug> <version>
+=item found <bug> [<version>]
 
 Indicate that a bug was found to exist in a particular package version.
 
@@ -738,7 +802,11 @@ Indicate that a bug was found to exist in a particular package version.
 
 sub bts_found {
     my $bug=checkbug(shift) or die "bts found: found what bug?\n";
-    my $version=shift or die "bts found: found \#$bug in what version?\n";
+    my $version=shift;
+    if (! defined $version) {
+	warn "bts: found has no version number, but sending to the BTS anyway\n";
+	$version="";
+    }
     opts_done(@_);
     mailbts("found $bug in $version", "found $bug $version");
 }
@@ -785,6 +853,10 @@ Note that a bug is no longer blocked from being fixed by a set of other bugs.
 
 sub bts_unblock {
     my $bug=checkbug(shift) or die "bts unblock: what bug is blocked?\n";
+    my $word=shift;
+    if ($word ne 'by' && $word ne 'with') {
+	    unshift @_, $word;
+    }
     my @blockers;
     foreach (@_) {
 	my $blocker=checkbug($_) or die "bts unblock: some blocking bug number(s) not valid\n";
@@ -1025,6 +1097,57 @@ sub bts_noowner {
     mailbts("bug $bug has no owner", "noowner $bug");
 }
 
+=item subscribe <bug> <email>
+
+Subscribe the given email address to the specified bug report.  If no email
+address is specified, the environment variable DEBEMAIL or EMAIL (in that
+order) is used.  If those are not set, or `!' is given as email address,
+your default address will be used.
+
+After executing this command, you will be sent a subscription confirmation to
+which you have to reply.  When subscribed to a bug report, you receive all
+relevant emails and notifications.  Use the unsubscribe command to unsubscribe.
+
+=cut
+
+sub bts_subscribe {
+    my $bug=checkbug(shift) or die "bts subscribe: subscribe to what bug?\n";
+    my $email=lc(shift);
+    if (defined $email and $email eq '!') { $email = undef; }
+    else {
+	$email ||= $ENV{'DEBEMAIL'};
+	$email ||= $ENV{'EMAIL'};
+    }
+    opts_done(@_);
+    mailto('subscription request for bug #' . $bug, '',
+	   $bug . '-subscribe@' . $btsserver, $email);
+}
+
+=item unsubscribe <bug> <email>
+
+Unsubscribe the given email address from the specified bug report.  As with
+subscribe above, if no email address is specified, the environment variables
+DEBEMAIL or EMAIL (in that order) is used.  If those are not set, or `!' is
+given as email address, your default address will be used.
+
+After executing this command, you will be sent an unsubscription confirmation
+to which you have to reply. Use the subscribe command to, well, subscribe.
+
+=cut
+
+sub bts_unsubscribe {
+    my $bug=checkbug(shift) or die "bts unsubscribe: unsubscribe from what bug?\n";
+    my $email=lc(shift);
+    if (defined $email and $email eq '!') { $email = undef; }
+    else {
+	$email ||= $ENV{'DEBEMAIL'};
+	$email ||= $ENV{'EMAIL'};
+    }
+    opts_done(@_);
+    mailto('unsubscription request for bug #' . $bug, '',
+	   $bug . '-unsubscribe@' . $btsserver, $email);
+}
+
 =item reportspam <bug>
 
 The reportspam command allows you to report a bug report as containing spam.
@@ -1047,8 +1170,20 @@ sub bts_reportspam {
     }
 }
 
+=item spamreport <bug>
+
+spamreport is a synonym for reportspam.
+
+=cut
+
+sub bts_spamreport {
+    goto &bts_reportspam;
+}
+
 =item cache [options] [<maint email> | <pkg> | src:<pkg> | from:<submitter>]
 
+=item cache [options] [release-critical | release-critical/... | RC]
+
 Generate or update a cache of bug reports for the given email address
 or package. By default it downloads all bugs belonging to the email
 address in the DEBEMAIL environment variable (or the EMAIL environment
@@ -1089,6 +1224,10 @@ Finally, -q or --quiet will suppress messages about caches being
 up-to-date, and giving the option twice will suppress all cache
 messages (except for error messages).
 
+Beware of caching RC, though: it will take a LONG time!  (With 1000+
+RC bugs and a delay of 5 seconds between bugs, you're looking at a
+minimum of 1.5 hours, and probably significantly more than that.)
+
 =cut
 
 sub bts_cache {
@@ -1133,12 +1272,12 @@ sub bts_cache {
     download("css/bugs.css");
 
     my $tocache;
-    if (@_ > 0) { $tocache=shift; }
+    if (@_ > 0) { $tocache=sanitizething(shift); }
     else { $tocache=''; }
     
     if (! length $tocache) {
-	$tocache=$ENV{'DEBEMAIL'} || $ENV{'EMAIL'};
-	if ($tocache =~ /^.*\s+<(.*)>$/) { $tocache = $1; }
+	$tocache=$ENV{'DEBEMAIL'} || $ENV{'EMAIL'} || '';
+	if ($tocache =~ /^.*\s<(.*)>\s*$/) { $tocache = $1; }
     }
     if (! length $tocache) {
 	die "bts cache: cache what?\n";
@@ -1147,7 +1286,7 @@ sub bts_cache {
     my @oldbugs = bugs_from_thing($tocache);
     
     # download index
-    download($tocache, 1);
+    download($tocache, '', 1);
 
     my %bugs = map { $_ => 1 } bugs_from_thing($tocache);
 
@@ -1171,7 +1310,7 @@ sub bts_cache {
     my $bugcount = 1;
     my $bugtotal = scalar keys %bugs;
     foreach my $bug (keys %bugs) {
-	download($bug, 1, 0, $bugcount, $bugtotal);
+	download($bug, '', 1, 0, $bugcount, $bugtotal);
         sleep $opt_cachedelay;
 	$bugcount++;
     }
@@ -1188,7 +1327,7 @@ sub bts_cache {
 
 =item cleancache <package> | src:<package> | <maintainer>
 
-=item cleancache from:<submitter> | tag:<tag> | <number> | ALL
+=item cleancache from:<submitter> | tag:<tag> | usertag:<tag> | <number> | ALL
 
 Clean the cache for the specified package, maintainer, etc., as
 described above for the "bugs" command, or clean the entire cache if
@@ -1201,7 +1340,7 @@ value of DEBEMAIL or EMAIL.
 
 sub bts_cleancache {
     prunecache();
-    my $toclean=shift;
+    my $toclean=sanitizething(shift);
     if (! defined $toclean) {
 	die "bts cleancache: clean what?\n";
     }
@@ -1292,6 +1431,7 @@ Valid options are:
    -f, --force-refresh    Reload all bug reports being cached, even unchanged
                           ones
    --no-force-refresh     Don\'t do so (default)
+   --sendmail=cmd         Sendmail command to use (default /usr/sbin/sendmail)
    --help, -h             Display this message
    --version, -v          Display version and copyright info
 
@@ -1311,6 +1451,19 @@ EOF
     }
 }
 
+# Strips any leading # or Bug# and trailing : from a thing if what's left is
+# a pure positive number;
+# also RC is a synonym for release-critical/other/all.html
+sub sanitizething {
+    my $bug = $_[0];
+    defined $bug or return undef;
+
+    return 'release-critical/other/all.html' if $bug eq 'RC';
+    return 'release-critical/index.html' if $bug eq 'release-critical';
+    $bug =~ s/^(?:(?:Bug)?\#)?(\d+):?$/$1/;
+    return $bug;
+}
+
 # Validate a bug number. Strips out extraneous leading junk, allowing
 # for things like "#74041" and "Bug#94921"
 sub checkbug {
@@ -1326,8 +1479,7 @@ sub checkbug {
 	}
     }
     
-    $bug=~s/^[^-0-9]*//;
-    $bug=~s/:$//;
+    $bug=~s/^(?:(?:bug)?\#)?(-?\d+):?$/$1/i;
     if (! exists $clonedbugs{$bug} &&
 	(! length $bug || $bug !~ /^[0-9]+$/)) {
 	warn "\"$_[0]\" does not look like a bug number\n" unless $quiet;
@@ -1364,9 +1516,15 @@ sub mailbtsall {
 	# We need to fake the From: line
 	my ($email, $name);
 	if (exists $ENV{'DEBFULLNAME'}) { $name = $ENV{'DEBFULLNAME'}; }
-	if (exists $ENV{'DEBEMAIL'}) { $email = $ENV{'DEBEMAIL'}; }
+	if (exists $ENV{'DEBEMAIL'}) {
+	    $email = $ENV{'DEBEMAIL'};
+	    if ($email =~ /^(.*?)\s+<(.*)>\s*$/) {
+		$name ||= $1;
+		$email = $2;
+	    }
+	}
 	if (exists $ENV{'EMAIL'}) {
-	    if ($ENV{'EMAIL'} =~ /^(.*)\s+<(.*)>$/) {
+	    if ($ENV{'EMAIL'} =~ /^(.*?)\s+<(.*)>\s*$/) {
 		$name ||= $1;
 		$email ||= $2;
 	    } else {
@@ -1391,6 +1549,7 @@ sub mailbtsall {
 	if (! defined $pid) {
 	    die "bts: Couldn't fork: $!\n";
 	}
+	$SIG{'PIPE'} = sub { die "bts: pipe for $sendmailcmd broke\n"; };
 	if ($pid) {
 	    # parent
 	    print MAIL <<"EOM";
@@ -1411,16 +1570,20 @@ EOM
 		exec("/bin/cat")
 		    or die "bts: error running cat: $!\n";
 	    } else {
-		exec("/usr/sbin/sendmail", "-t")
+		exec(split(' ', $sendmailcmd), "-t")
 		    or die "bts: error running sendmail: $!\n";
 	    }
 	}
     }
     else {  # No DEBEMAIL
 	unless (system("command -v mail >/dev/null 2>&1") == 0) {
-	    die "bts: You need to either set DEBEMAIL or have the mailx package to do this!\n";
+	    die "bts: You need to either set DEBEMAIL or have the mailx/mailutils package\ninstalled to send mail!\n";
 	}
 	my $pid = open(MAIL, "|-");
+	if (! defined $pid) {
+	    die "bts: Couldn't fork: $!\n";
+	}
+	$SIG{'PIPE'} = sub { die "bts: pipe for mail broke\n"; };
 	if ($pid) {
 	    # parent
 	    print MAIL $body;
@@ -1432,7 +1595,73 @@ EOM
 		exec("/bin/cat")
 		    or die "bts: error running cat: $!\n";
 	    } else {
-		exec("mail", "-s$subject", $btsemail)
+		exec("mail", "-s", $subject, $btsemail)
+		    or die "bts: error running mail: $!\n";
+	    }
+	}
+    }
+}
+
+# A simplified version of mailbtsall which sends one message only to
+# a specified address using the specified email From: header
+sub mailto {
+    my ($subject, $body, $to, $from) = @_;
+
+    if (defined $from) {
+	my $date = `822-date`;
+	chomp $date;
+
+	my $pid = open(MAIL, "|-");
+	if (! defined $pid) {
+	    die "bts: Couldn't fork: $!\n";
+	}
+	$SIG{'PIPE'} = sub { die "bts: pipe for $sendmailcmd broke\n"; };
+	if ($pid) {
+	    # parent
+	    print MAIL <<"EOM";
+From: $from
+To: $to
+Subject: $subject
+Date: $date
+X-BTS-Version: $version
+
+# Automatically generated email from bts, devscripts version $version
+$body
+EOM
+	    close MAIL or die "bts: sendmail error: $!\n";
+	}
+	else {
+	    # child
+	    if ($debug) {
+		exec("/bin/cat")
+		    or die "bts: error running cat: $!\n";
+	    } else {
+		exec(split(' ', $sendmailcmd), "-t")
+		    or die "bts: error running sendmail: $!\n";
+	    }
+	}
+    }
+    else {  # No $from
+	unless (system("command -v mail >/dev/null 2>&1") == 0) {
+	    die "bts: You need to either specify an email address (say using DEBEMAIL)\n or have the mailx/mailutils package installed to send mail!\n";
+	}
+	my $pid = open(MAIL, "|-");
+	if (! defined $pid) {
+	    die "bts: Couldn't fork: $!\n";
+	}
+	$SIG{'PIPE'} = sub { die "bts: pipe for mail broke\n"; };
+	if ($pid) {
+	    # parent
+	    print MAIL $body;
+	    close MAIL or die "bts: mail: $!\n";
+	}
+	else {
+	    # child
+	    if ($debug) {
+		exec("/bin/cat")
+		    or die "bts: error running cat: $!\n";
+	    } else {
+		exec("mail", "-s", $subject, $to)
 		    or die "bts: error running mail: $!\n";
 	    }
 	}
@@ -1554,6 +1783,7 @@ sub fold_from_header {
 # live thing as a (non-empty) string
 sub download {
     my $thing=shift;
+    my $thgopts=shift ||'';
     my $manual=shift;  # true="bts cache", false="bts show/bug"
     my $mboxing=shift;  # true="bts --mbox show/bugs", and only if $manual=0
     my $bug_current=shift;  # current bug being downloaded if caching
@@ -1563,13 +1793,9 @@ sub download {
     my $url;
 
     # What URL are we to download?
-    if ($thing =~ /;/) {
+    if ($thgopts ne '') {
 	# have to be intelligent here :/
-	if ($thing =~ /^\d+;/) {
-	    $url = "$btscgibugurl?bug=$thing";
-	} else {
-	    $url = "$btscgipkgurl?pkg=$thing";
-	}
+	$url = thing_to_url($thing) . $thgopts;
     } else {
 	# let the BTS be intelligent
 	$url = "$btsurl$thing";
@@ -1581,8 +1807,8 @@ sub download {
 
     chdir($cachedir) || die "bts: chdir $cachedir: $!\n";
 
-    if (-f cachefile($thing)) {
-	($timestamp, $versionstamp) = get_timestamp($thing);
+    if (-f cachefile($thing, $thgopts)) {
+	($timestamp, $versionstamp) = get_timestamp($thing, $thgopts);
 	$timestamp ||= 0;
 	$versionstamp ||= 0;
 	# And ensure we preserve any manual setting
@@ -1616,7 +1842,7 @@ sub download {
 	# we have an up-to-date version already, nothing to do
 	# and $timestamp is guaranteed to be well-defined
 	if (is_automatic($timestamp) and $manual) {
-	    set_timestamp($thing, make_manual($timestamp), $versionstamp);
+	    set_timestamp($thing, $thgopts, make_manual($timestamp), $versionstamp);
 	}
 
 	if (! $quiet and $manual and $thing ne "css/bugs.css") {
@@ -1643,14 +1869,14 @@ sub download {
 	}
 
 	my $data = $livepage;  # work on a copy, not the original
-	my $cachefile=cachefile($thing);
+	my $cachefile=cachefile($thing,$thgopts);
 	open (OUT_CACHE, ">$cachefile") or die "bts: open $cachefile: $!\n";
 
 	$data = mangle_cache_file($data, $thing, $bug2filename, $timestamp);
 	print OUT_CACHE $data;
 	close OUT_CACHE or die "bts: problems writing to $cachefile: $!\n";
 
-	set_timestamp($thing,
+	set_timestamp($thing, $thgopts,
 	    $manual ? make_manual($timestamp) : make_automatic($timestamp),
 	    $version);
 
@@ -1674,19 +1900,19 @@ sub download_attachments {
     my ($thing, $toppage, $timestamp) = @_;
     my %bug2filename;
 
-    # We search for appropriate strings in the toppage, and save the
+    # We search for appropriate strings in the top page, and save the
     # attachments in files with names as follows:
     # - if the attachment specifies a filename, save as bug#/msg#-att#/filename
     # - if not, save as bug#/msg#-att# with suffix .txt if plain/text and
     #   .html if plain/html, no suffix otherwise (too much like hard work!)
     # Since messages are never modified retrospectively, we don't download
-    # attachements which have already been downloaded
+    # attachments which have already been downloaded
     
     # Yuck, yuck, yuck.  This regex splits the $data string at every
     # occurrence of either "[<a " or plain "<a ", preserving any "[".
     my @data = split /(?:(?=\[<[Aa]\s)|(?<!\[)(?=<[Aa]\s))/, $toppage;
     foreach (@data) {
-	next unless m%<a href="(bugreport\.cgi[^\"]+)">%;
+	next unless m%<a(?: class=\".*?\")? href="(bugreport\.cgi[^\"]+)">%i;
 	my $ref = $1;
 	my ($msg, $filename) = href_to_filename($_);
 
@@ -1812,30 +2038,51 @@ sub mangle_cache_file {
     $data =~ s%(<BODY.*>)%$1<p><em>[Locally cached on $time by devscripts version $version]</em></p>%i;
     $data =~ s%href="/css/bugs.css"%href="bugs.css"%;
 
-    # Yuck, yuck, yuck.  This regex splits the $data string at every
-    # occurrence of either "[<a " or plain "<a ", preserving any "[".
-    my @data = split /(?:(?=\[<[Aa]\s)|(?<!\[)(?=<[Aa]\s))/, $data;
-    foreach (@data) {
-	if (m%<a href=\"bugreport\.cgi[^\?]*\?bug=(\d+)%i) {
-	    my $bug = $1;
-	    my ($msg, $filename) = href_to_filename($_);
-
-	    if ($bug eq $thing and defined $msg) {
-		if ($fullmode or
-		    (! $fullmode and exists $$bug2filename{$msg})) {
-		    s%<a href="(bugreport\.cgi[^\"]*)">(.+?)</a>%<a href="$filename">$2</a> (<a href="$btscgiurl$1">online</a>)%i;
-		} else {
-		    s%<a href="(bugreport\.cgi[^\"]*)">(.+?)</a>%$2 (<a href="$btscgiurl$1">online</a>)%i;
-		}
-	    } else {
-		s%<a href="(bugreport\.cgi[^\?]*\?bug=(\d+)[^\"]*)">(.+?)</a>%<a href="$2.html">$3</a> (<a href="$btscgiurl$1">online</a>)%i;
+    my @data;
+    # We have to distinguish between release-critical pages and normal BTS
+    # pages as they have a different structure
+    if ($thing =~ /^release-critical/) {
+	@data = split /(?=<[Aa])/, $data;
+	foreach (@data) {
+	    s%<a href="(http://bugs.debian.org/cgi-bin/bugreport\.cgi.*bug=(\d+)[^\"]*)">(.+?)</a>%<a href="$2.html">$3</a> (<a href="$1">online</a>)%i;
+	    s%<a href="(http://bugs.debian.org/cgi-bin/pkgreport\.cgi.*pkg=([^\"&;]+)[^\"]*)">(.+?)</a>%<a href="$2.html">$3</a> (<a href="$1">online</a>)%i;
+	    # References to other bug lists on bugs.d.o/release-critical
+	    if (m%<a href="((?:debian|other)[-a-z/]+\.html)"%i) {
+		my $ref = 'release-critical/'.$1;
+		$ref =~ s%/%_%g;
+		s%<a href="((?:debian|other)[-a-z/]+\.html)">(.+?)</a>%<a href="$ref">$2</a> (<a href="${btsurl}release-critical/$1">online</a>)%i;
 	    }
+	    # Maintainer email address - YUCK!!
+	    s%<a href="(http://bugs.debian.org/([^\"?]*\@[^\"?]*))">(.+?)</a>&gt;%<a href="$2.html">$3</a>&gt; (<a href="$1">online</a>)%i;
+	    # Graph - we don't download
+	    s%<img src="graph.png" alt="Graph of RC bugs">%<img src="${btsurl}release-critical/graph.png" alt="Graph of RC bugs (online)">%
 	}
-	else {
-	    s%<a href="(pkgreport\.cgi\?(?:pkg|maint)=([^\"&;]+)[^\"]*)">(.+?)</a>%<a href="$2.html">$3</a> (<a href="$btscgiurl$1">online</a>)%i;
-	    s%<a href="(pkgreport\.cgi\?src=([^\"&;]+)[^\"]*)">(.+?)</a>%<a href="src_$2.html">$3</a> (<a href="$btscgiurl$1">online</a>)%i;
-	    s%<a href="(pkgreport\.cgi\?submitter=([^\"&;]+)[^\"]*)">(.+?)</a>%<a href="from_$2.html">$3</a> (<a href="$btscgiurl$1">online</a>)%i;
-	    s%<a href="(?:/cgi-bin/)?(bugspam\.cgi[^\"]+)">%<a href="$btscgiurl$1">%i;
+    } else {
+	# Yuck, yuck, yuck.  This regex splits the $data string at every
+	# occurrence of either "[<a " or plain "<a ", preserving any "[".
+	@data = split /(?:(?=\[<[Aa]\s)|(?<!\[)(?=<[Aa]\s))/, $data;
+	foreach (@data) {
+	    if (m%<a(?: class=\".*?\")? href=\"bugreport\.cgi[^\?]*\?bug=(\d+)%i) {
+		my $bug = $1;
+		my ($msg, $filename) = href_to_filename($_);
+
+		if ($bug eq $thing and defined $msg) {
+		    if ($fullmode or
+			(! $fullmode and exists $$bug2filename{$msg})) {
+			s%<a((?: class=\".*?\")?) href="(bugreport\.cgi[^\"]*)">(.+?)</a>%<a$1 href="$filename">$3</a> (<a$1 href="$btscgiurl$2">online</a>)%i;
+		    } else {
+			s%<a((?: class=\".*?\")?) href="(bugreport\.cgi[^\"]*)">(.+?)</a>%$3 (<a$1 href="$btscgiurl$2">online</a>)%i;
+		    }
+		} else {
+		    s%<a((?: class=\".*?\")?) href="(bugreport\.cgi[^\?]*\?bug=(\d+)[^\"]*)">(.+?)</a>%<a$1 href="$3.html">$4</a> (<a$1 href="$btscgiurl$2">online</a>)%i;
+		}
+	    }
+	    else {
+		s%<a((?: class=\".*?\")?) href="(pkgreport\.cgi\?(?:pkg|maint)=([^\"&;]+)[^\"]*)">(.+?)</a>%<a$1 href="$3.html">$4</a> (<a$1 href="$btscgiurl$2">online</a>)%i;
+		s%<a((?: class=\".*?\")?) href="(pkgreport\.cgi\?src=([^\"&;]+)[^\"]*)">(.+?)</a>%<a$1 href="src_$3.html">$4</a> (<a$1 href="$btscgiurl$2">online</a>)%i;
+		s%<a((?: class=\".*?\")?) href="(pkgreport\.cgi\?submitter=([^\"&;]+)[^\"]*)">(.+?)</a>%<a$1 href="from_$3.html">$4</a> (<a$1 href="$btscgiurl$2">online</a>)%i;
+		s%<a((?: class=\".*?\")?) href="(?:/cgi-bin/)?(bugspam\.cgi[^\"]+)">%<a$1 href="$btscgiurl$2">%i;
+	    }
 	}
     }
 
@@ -1846,13 +2093,14 @@ sub mangle_cache_file {
 # Removes a specified thing from the cache
 sub deletecache {
     my $thing=shift;
+    my $thgopts=shift || '';
 
     if (! -d $cachedir) {
 	die "bts: deletecache() called but no cachedir!\n";
     }
 
-    delete_timestamp($thing);
-    unlink cachefile($thing);
+    delete_timestamp($thing,$thgopts);
+    unlink cachefile($thing,$thgopts);
     if ($thing =~ /^\d+$/) {
 	rmtree("$cachedir/$thing", 0, 1) if -d "$cachedir/$thing";
 	unlink("$cachedir/$thing.mbox") if -f "$cachedir/$thing.mbox";
@@ -1862,12 +2110,18 @@ sub deletecache {
 # Given a thing, returns the filename for it in the cache.
 sub cachefile {
     my $thing=shift;
+    my $thgopts=shift || '';
     if ($thing eq '') { die "bts: cachefile given empty argument\n"; }
     if ($thing =~ /bugs.css$/) { return $cachedir."bugs.css" }
     $thing =~ s/^src:/src_/;
     $thing =~ s/^from:/from_/;
     $thing =~ s/^tag:/tag_/;
-    return $cachedir.$thing.".html";
+    $thing =~ s/^usertag:/usertag_/;
+    $thing =~ s%^release-critical/index\.html$%release-critical.html%;
+    $thing =~ s%/%_%g;
+    $thgopts =~ s/;/_3B/g;
+    $thgopts =~ s/=/_3D/g;
+    return $cachedir.$thing.$thgopts.($thing =~ /\.html$/ ? "" : ".html");
 }
 
 # Given a thing, returns the filename for its mbox in the cache.
@@ -1887,17 +2141,59 @@ sub cachebugdir {
 # "thing".
 sub cachefile_to_thing {
     my $thing=basename(shift, '.html');
+    my $thgopts='';
     $thing =~ s/^src_/src:/;
     $thing =~ s/^from_/from:/;
     $thing =~ s/^tag_/tag:/;
-    return $thing;
+    $thing =~ s/^usertag_/usertag:/;
+    $thing =~ s%^release-critical\.html$%release-critical/index\.html%;
+    $thing =~ s%_%/%g;
+    $thing =~ s/_3B/;/g;
+    $thing =~ s/_3D/=/g;
+    $thing =~ /^(.*?)((?:;.*)?)$/;
+    ($thing, $thgopts) = ($1, $2);
+    return ($thing, $thgopts);
+}
+
+# Given a thing, gives the official BTS cgi page for it
+sub thing_to_url {
+    my $thing = shift;
+    my $thingurl;
+
+    # have to be intelligent here :/
+    if ($thing =~ /^\d+$/) {
+	$thingurl = $btscgibugurl."?bug=".$thing;
+    } elsif ($thing =~ /^from:/) {
+	($thingurl = $thing) =~ s/^from:/submitter=/;
+	$thingurl = $btscgipkgurl.'?'.$thingurl;
+    } elsif ($thing =~ /^src:/) {
+	($thingurl = $thing) =~ s/^src:/src=/;
+	$thingurl = $btscgipkgurl.'?'.$thingurl;
+    } elsif ($thing =~ /^tag:/) {
+	($thingurl = $thing) =~ s/^tag:/tag=/;
+	$thingurl = $btscgipkgurl.'?'.$thingurl;
+    } elsif ($thing =~ /^usertag:/) {
+	($thingurl = $thing) =~ s/^usertag:/usertag=/;
+	$thingurl = $btscgipkgurl.'?'.$thingurl;
+    } elsif ($thing =~ m%^release-critical(\.html|/(index\.html)?)?$%) {
+	$thingurl = $btsurl . 'release-critical/index.html';
+    } elsif ($thing =~ m%^release-critical/%) {
+	$thingurl = $btsurl . $thing;
+    } elsif ($thing =~ /\@/) { # so presume it's a maint request
+	$thingurl = $btscgipkgurl.'?maint='.$thing;
+    } else { # it's a package, or had better be...
+	$thingurl = $btscgipkgurl.'?pkg='.$thing;
+    }
+
+    return $thingurl;
 }
 
 # Given a thing, reads all links to bugs from the corresponding cache file
 # if there is one, and returns a list of them.
 sub bugs_from_thing {
     my $thing=shift;
-    my $cachefile=cachefile($thing);
+    my $thgopts=shift || '';
+    my $cachefile=cachefile($thing,$thgopts);
 
     if (-f $cachefile) {
 	local $/;
@@ -1906,7 +2202,9 @@ sub bugs_from_thing {
 	close IN;
 
 	return $data =~ m!href="(\d+)\.html"!g;
-    } else { return (); }
+    } else {
+	return ();
+    }
 }
 
 # Given an <a href="bugreport.cgi?...>...</a> string, return a
@@ -1915,7 +2213,7 @@ sub href_to_filename {
     my $href = $_[0];
     my ($msg, $filename);
 
-    if ($href =~ m%\[<a href="bugreport\.cgi([^\?]*)\?bug=(\d+)([^\"]*)">.*?\(([^,]*), .*?\)\]%) {
+    if ($href =~ m%\[<a(?: class=\".*?\")? href="bugreport\.cgi([^\?]*)\?bug=(\d+)([^\"]*)">.*?\(([^,]*), .*?\)\]%) {
 	# this looks like an attachment; $1 should give the MIME-type
 	my $urlfilename = $1;
 	my $bug = $2;
@@ -1940,11 +2238,12 @@ sub href_to_filename {
 	    $filename = "$bug/$msg$fileext";
 	}
     }
-    elsif ($href =~ m%<a href="bugreport\.cgi([^\?]*)\?bug=(\d+)([^\"]*)">%) {
+    elsif ($href =~ m%<a(?: class=\".*?\")? href="bugreport\.cgi([^\?]*)\?bug=(\d+)([^\"]*)">%) {
 	my $urlfilename = $1;
 	my $bug = $2;
 	my $ref = $3;
 	$ref =~ s/&(?:amp;)?/;/g;  # normalise all hrefs
+	$ref =~ s/;archive=(yes|no)\b//;
 
 	if ($ref =~ /;msg=(\d+)$/) {
 	    $msg = $1;
@@ -1974,25 +2273,30 @@ sub href_to_filename {
     return ($msg, $filename);
 }
 
-# Browses a given thing, with possible caching.
+# Browses a given thing, with preprocessed list of URL options such as
+# ";opt1=val1;opt2=val2" with possible caching if there are no options
 sub browse {
     prunecache();
     my $thing=shift;
+    my $thgopts=shift || '';
     
     if ($thing eq '') {
+	if ($thgopts ne '') {
+	    die "bts: you can only give options for a BTS page if you specify a bug/maint/... .\n";
+	}
 	runbrowser($btsurl);
 	return;
     }
 
     my $hascache=-d $cachedir;
-    my $cachefile=cachefile($thing);
+    my $cachefile=cachefile($thing,$thgopts);
     my $mboxfile=mboxfile($thing);
     if ($mboxmode and ! $mboxfile) {
 	die "bts: you can only request a mailbox for a single bug report.\n";
     }
 
     # Check that if we're requesting a tag, that it's a valid tag
-    if ($thing =~ /(?:^|;)tag[:=]([^;]*)/) {
+    if (($thing.$thgopts) =~ /(?:^|;)(?:tag|include|exclude)[:=]([^;]*)/) {
 	unless (exists $valid_tags{$1}) {
 	    die "bts: invalid tag requested: $1\nRecognised tag names are: " . join(" ", @valid_tags) . "\n";
 	}
@@ -2014,7 +2318,7 @@ sub browse {
     }
     # else we're in online mode
     elsif ($hascache && $caching && have_lwp() && $thing ne '') {
-	my $live=download($thing, 0, $mboxmode);
+	my $live=download($thing, $thgopts, 0, $mboxmode);
 	
 	if ($mboxmode) {
 	    runmailreader($mboxfile);
@@ -2047,13 +2351,9 @@ sub browse {
 	    my ($fh, $fn) = download_mbox($thing, 1);
 	    runmailreader($fn);
 	} else {
-	    if ($thing =~ /;/) {
-		# have to be intelligent here :/
-		if ($thing =~ /^\d+;/) {
-		    runbrowser($btscgibugurl."?bug=".$thing);
-		} else {
-		    runbrowser($btscgipkgurl."?pkg=".$thing);
-		}
+	    if ($thgopts ne '') {
+		my $thingurl = thing_to_url($thing);		
+		runbrowser($thingurl.$thgopts);
 	    } else {
 		# let the BTS be intelligent
 		runbrowser($btsurl.$thing);
@@ -2114,15 +2414,15 @@ sub prunecache {
 
     my @unrecognised;
     foreach my $oldfile (@oldfiles) {
-	my $thing=cachefile_to_thing($oldfile);
-	unless (exists $timestamp{$thing}) {
+	my ($thing, $thgopts) = cachefile_to_thing($oldfile);
+	unless (defined get_timestamp($thing, $thgopts)) {
 	    push @unrecognised, $oldfile;
 	    next;
 	}
-	next if is_manual(get_timestamp($thing));
+	next if is_manual(get_timestamp($thing, $thgopts));
 	
 	# Otherwise, it's automatic and we purge it
-	deletecache($thing);
+	deletecache($thing, $thgopts);
     }
 
     untie %timestamp;
@@ -2165,19 +2465,20 @@ sub runmailreader {
 
 sub get_timestamp {
     my $thing = shift;
+    my $thgopts = shift || '';
     my $timestamp = undef;
     my $versionstamp = undef;
 
     if (tied %timestamp) {
-	($timestamp, $versionstamp) = split /;/, $timestamp{$thing}
-	    if exists $timestamp{$thing};
+	($timestamp, $versionstamp) = split /;/, $timestamp{$thing.$thgopts}
+	    if exists $timestamp{$thing.$thgopts};
     } else {
 	tie (%timestamp, "Devscripts::DB_File_Lock", $timestampdb,
 	     O_RDONLY(), 0600, $DB_HASH, "read")
 	    or die "bts: couldn't open DB file $timestampdb for reading: $!\n";
 
-	($timestamp, $versionstamp) = split /;/, $timestamp{$thing}
-	    if exists $timestamp{$thing};
+	($timestamp, $versionstamp) = split /;/, $timestamp{$thing.$thgopts}
+	    if exists $timestamp{$thing.$thgopts};
 
 	untie %timestamp;
     }
@@ -2187,17 +2488,18 @@ sub get_timestamp {
 
 sub set_timestamp {
     my $thing = shift;
+    my $thgopts = shift || '';
     my $timestamp = shift;
     my $versionstamp = shift || $version;
 
     if (tied %timestamp) {
-	$timestamp{$thing} = "$timestamp;$versionstamp";
+	$timestamp{$thing.$thgopts} = "$timestamp;$versionstamp";
     } else {
 	tie (%timestamp, "Devscripts::DB_File_Lock", $timestampdb,
 	     O_RDWR()|O_CREAT(), 0600, $DB_HASH, "write")
 	    or die "bts: couldn't open DB file $timestampdb for writing: $!\n";
 
-	$timestamp{$thing} = "$timestamp;$versionstamp";
+	$timestamp{$thing.$thgopts} = "$timestamp;$versionstamp";
 
 	untie %timestamp;
     }
@@ -2205,15 +2507,16 @@ sub set_timestamp {
 
 sub delete_timestamp {
     my $thing = shift;
+    my $thgopts = shift || '';
 
     if (tied %timestamp) {
-	delete $timestamp{$thing};
+	delete $timestamp{$thing.$thgopts};
     } else {
 	tie (%timestamp, "Devscripts::DB_File_Lock", $timestampdb,
 	     O_RDWR()|O_CREAT(), 0600, $DB_HASH, "write")
 	    or die "bts: couldn't open DB file $timestampdb for writing: $!\n";
 
-	delete $timestamp{$thing};
+	delete $timestamp{$thing.$thgopts};
 
 	untie %timestamp;
     }
@@ -2392,6 +2695,16 @@ command line parameter being used.  Only has an effect on the cache
 command.  The default is I<no>.  See the cache command for more
 information.
 
+=item BTS_MAIL_READER
+
+If this is set, specifies a mail reader to use instead of mutt.  Same as
+the --mailreader command line option.
+
+=item BTS_SENDMAIL_COMMAND
+
+If this is set, specifies a sendmail command to use instead of
+/usr/sbin/sendmail.  Same as the --sendmail command line option.
+
 =cut
 
 =head1 COPYRIGHT
