@@ -4,11 +4,11 @@
 # Copyright 1998,1999 Yann Dirson <dirson@debian.org>
 # Perl version:
 # Copyright 1999,2000,2001 by Julian Gilbey <jdg@debian.org>
-# 
+#
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License, version 2 ONLY,
 # as published by the Free Software Foundation.
-# 
+#
 # This program is distributed in the hope that it will be useful,
 # but WITHOUT ANY WARRANTY; without even the implied warranty of
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
@@ -22,6 +22,7 @@ use File::Temp qw/ tempdir /;
 
 # Predeclare functions
 sub process_debc($$);
+sub process_debI($);
 sub mktmpdirs();
 sub fatal(@);
 
@@ -49,10 +50,12 @@ Valid options are:
                             (multiple permitted), using regexp substitution
    --dirs, -d             Note changes in directories as well as files
    --nodirs               Do not note changes in directories (default)
-   --nocontrol            Skip comparing control files when comparing
-                            two .debs
-   --control              Do compare control files when comparing
-                            two .debs (default)
+   --nocontrol            Skip comparing control files
+   --control              Do compare control files
+   --controlfiles FILE,FILE,...
+                          Which control files to compare; default is just
+                            control; could include preinst, etc, config or
+                            ALL to compare all control files present
    --wp, --wl, --wt       Pass the option -p, -l, -t respectively to wdiff
                             (only one should be used)
    --show-moved           Indicate also all files which have moved
@@ -63,6 +66,7 @@ Valid options are:
                             renamed TO; only of interest with --show-moved
                             (multiple permitted)
    --quiet, -q            Be quiet if no differences were found
+   --exclude PATTERN      Exclude files that match PATTERN
 
 Default settings modified by devscripts configuration files:
 $modified_conf_msg
@@ -83,6 +87,7 @@ EOF
 
 my $ignore_dirs = 1;
 my $compare_control = 1;
+my $controlfiles = 'control';
 my $show_moved = 0;
 my $wdiff_opt = '';
 
@@ -99,11 +104,12 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
     my %config_vars = (
 		       'DEBDIFF_DIRS' => 'no',
 		       'DEBDIFF_CONTROL' => 'yes',
+		       'DEBDIFF_CONTROLFILES' => 'control',
 		       'DEBDIFF_SHOW_MOVED' => 'no',
 		       'DEBDIFF_WDIFF_OPT' => '',
 		       );
     my %config_default = %config_vars;
-    
+
     my $shell_cmd;
     # Set defaults
     foreach my $var (keys %config_vars) {
@@ -134,24 +140,25 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
 
     $ignore_dirs = $config_vars{'DEBDIFF_DIRS'} eq 'yes' ? 0 : 1;
     $compare_control = $config_vars{'DEBDIFF_CONTROL'} eq 'no' ? 0 : 1;
+    $controlfiles = $config_vars{'DEBDIFF_CONTROLFILES'};
     $show_moved = $config_vars{'DEBDIFF_SHOW_MOVED'} eq 'yes' ? 1 : 0;
     $wdiff_opt = $config_vars{'DEBDIFF_WDIFF_OPT'} =~ /^-([plt])$/ ? $1 : '';
 }
 
 # Are they a pair of debs, changes or dsc files, or a list of debs?
 my $type = '';
+my @excludes = ();
 my @move = ();
 my %renamed = ();
 
-##
-## handle command-line options
-##
+
+# handle command-line options
 
 while (@ARGV) {
     if ($ARGV[0] =~ /^(--help|-h)$/) { usage(); exit 0; }
     if ($ARGV[0] =~ /^(--version|-v)$/) { print $version; exit 0; }
     if ($ARGV[0] =~ /^(--move(-regex)?|-m)$/) {
-	fatal "Malformed command-line options; run $progname --help for more info"
+	fatal "Malformed command-line option $ARGV[0]; run $progname --help for more info"
 	    unless @ARGV >= 3;
 
 	my $regex = $ARGV[0] eq '--move-regex' ? 1 : 0;
@@ -174,7 +181,7 @@ while (@ARGV) {
 	push @move, [$regex, $from, $to];
     }
     elsif ($ARGV[0] eq '--renamed') {
-	fatal "Malformed command-line options; run $progname --help for more info"
+	fatal "Malformed command-line option $ARGV[0]; run $progname --help for more info"
 	    unless @ARGV >= 3;
 	shift @ARGV;
 
@@ -182,6 +189,28 @@ while (@ARGV) {
 	my $to   = shift;
 	$renamed{$from} = $to;
     }
+    elsif ($ARGV[0] eq '--exclude') {
+	fatal "Malformed command-line option $ARGV[0]; run $progname --help for more info"
+	    unless @ARGV >= 2;
+	shift @ARGV;
+
+	my $exclude = shift;
+	push @excludes, $exclude;
+    }
+    elsif ($ARGV[0] =~ s/^--exclude=//) {
+	my $exclude = shift;
+	push @excludes, $exclude;
+    }
+    elsif ($ARGV[0] eq '--controlfiles') {
+	fatal "Malformed command-line option $ARGV[0]; run $progname --help for more info"
+	    unless @ARGV >= 2;
+	shift @ARGV;
+
+	$controlfiles = shift;
+    }
+    elsif ($ARGV[0] =~ s/^--controlfiles=//) {
+	$controlfiles = shift;
+    }
     elsif ($ARGV[0] =~ /^(--dirs|-d)$/) { $ignore_dirs = 0; shift; }
     elsif ($ARGV[0] eq '--nodirs') { $ignore_dirs = 1; shift; }
     elsif ($ARGV[0] =~ /^(--quiet|-q)$/) { $quiet = 1; shift; }
@@ -233,8 +262,9 @@ if (! $type) {
 # %deb1 and %deb2, each key of which is a .deb name and each value is
 # a list ref.  Note we need to use our, not my, as we will be symbolically
 # referencing these variables
+my @CommonDebs = ();
 my @singledeb;
-our (%debs1, %debs2, %files1, %files2, @D1, @D2, $dir1, $dir2);
+our (%debs1, %debs2, %files1, %files2, @D1, @D2, $dir1, $dir2, %DebPaths1, %DebPaths2);
 
 if ($type eq 'deb') {
     no strict 'refs';
@@ -242,11 +272,14 @@ if ($type eq 'deb') {
 	my $deb = shift;
 	my $debc = `env LC_ALL=C dpkg-deb -c $deb`;
 	$? == 0 or fatal "dpkg-deb -c $deb failed!";
+	my $debI = `env LC_ALL=C dpkg-deb -I $deb`;
+	$? == 0 or fatal "dpkg-deb -I $deb failed!";
 	# Store the name for later
 	$singledeb[$i] = $deb;
 	# get package name itself
 	$deb =~ s,.*/,,; $deb =~ s/_.*//;
 	@{"D$i"} = @{process_debc($debc,$i)};
+	push @{"D$i"}, @{process_debI($debI)};
     }
 }
 elsif ($type eq 'changes' or $type eq 'debs') {
@@ -277,26 +310,25 @@ elsif ($type eq 'changes' or $type eq 'debs') {
 		last if $infiles and /^[^ ]/;
 		/^Files:/ and $infiles=1, next;
 		next unless $infiles;
-		/ (\S*.u?deb)$/ and push @debs, $1;
+		/ (\S*.u?deb)$/ and push @debs, dirname($changes) . '/' . $1;
 	    }
 	    close CHANGES
 		or fatal "Problem reading $changes: $!";
 
-	    chdir dirname($changes)
-		or fatal "Couldn't chdir ", dirname($changes), ": $!";
-
 	    # Is there only one .deb listed?
 	    if (@debs == 1) {
-		$singledeb[$i] = dirname($changes) . '/' . $debs[0];
+		$singledeb[$i] = $debs[0];
 	    }
 	}
 
-	my %D = ();
 	foreach my $deb (@debs) {
 	    no strict 'refs';
 	    fatal "Can't read file: $deb" unless -r $deb;
 	    my $debc = `env LC_ALL=C dpkg-deb -c $deb`;
 	    $? == 0 or fatal "dpkg-deb -c $deb failed!";
+	    my $debI = `env LC_ALL=C dpkg-deb -I $deb`;
+	    $? == 0 or fatal "dpkg-deb -I $deb failed!";
+	    my $debpath = $deb;
 	    # get package name itself
 	    $deb =~ s,.*/,,; $deb =~ s/_.*//;
 	    $deb = $renamed{$deb} if $i == 1 and exists $renamed{$deb};
@@ -305,13 +337,18 @@ elsif ($type eq 'changes' or $type eq 'debs') {
 	    } else {
 		${"debs$i"}{$deb} = 1;
 	    }
+	    ${"DebPaths$i"}{$deb} = $debpath;
 	    foreach my $file (@{process_debc($debc,$i)}) {
+		${"files$i"}{$file} ||= ":";
 		${"files$i"}{$file} .= "$deb:";
-		${"D$i"}{$file} = 1;
+	    }
+	    foreach my $control (@{process_debI($debI)}) {
+		${"files$i"}{$control} ||= ":";
+		${"files$i"}{$control} .= "$deb:";
 	    }
 	}
 	no strict 'refs';
-	@{"D$i"} = keys %{"D$i"};
+	@{"D$i"} = keys %{"files$i"};
 	# Go back again
 	chdir $pwd or fatal "Couldn't chdir $pwd: $!";
     }
@@ -361,11 +398,11 @@ elsif ($type eq 'dsc') {
     }
 
     # Do we have interdiff?
-    system("(interdiff --version) >/dev/null 2>&1");
+    system("command -v interdiff >/dev/null 2>&1");
     my $use_interdiff = ($?==0) ? 1 : 0;
 
     if ($origs[1] eq $origs[2] and defined $diffs[1] and defined $diffs[2]
-	and $use_interdiff) {
+	and scalar(@excludes) == 0 and $use_interdiff) {
 	# same orig tar ball and interdiff exists
 	my $rv = system('interdiff', '-z', $diffs[1], $diffs[2]);
 	if ($rv) {
@@ -374,7 +411,8 @@ elsif ($type eq 'dsc') {
     } else {
 	# Any other situation
 	if ($origs[1] eq $origs[2] and
-	    defined $diffs[1] and defined $diffs[2]) {
+	    defined $diffs[1] and defined $diffs[2] and
+	    scalar(@excludes) == 0) {
 	    warn "Warning: You do not seem to have interdiff (in the patchutils package)\ninstalled; this program would use it if it were available.\n";
 	}
 	# possibly different orig tarballs, or no interdiff installed
@@ -393,7 +431,12 @@ elsif ($type eq 'dsc') {
 	    }
 	    closedir(DIR);
 	}
-	system ("diff", "-Nru", "$dir1/$sdir1", "$dir2/$sdir2");
+	my @command = ("diff", "-Nru");
+	for my $exclude (@excludes) {
+	    push @command, ("--exclude", $exclude);
+	}
+	push @command, ("$dir1/$sdir1", "$dir2/$sdir2");
+	system @command;
     }
 
     exit 0;
@@ -402,67 +445,59 @@ else {
     fatal "Internal error: \$type = $type unrecognised";
 }
 
-##
-## Compare
-##
 
-if ($show_moved and $type ne 'deb') {
-    # We first check the list of .debs
+# Compare
+# Start by a piece of common code to set up the @CommonDebs list and the like
+
+my (@deblosses, @debgains);
+
+{
     my %debs;
     grep $debs{$_}--, keys %debs1;
     grep $debs{$_}++, keys %debs2;
 
-    my @losses = sort grep $debs{$_} < 0, keys %debs;
-    my @gains  = sort grep $debs{$_} > 0, keys %debs;
+    my @deblosses = sort grep $debs{$_} < 0, keys %debs;
+    my @debgains  = sort grep $debs{$_} > 0, keys %debs;
+    @CommonDebs= sort grep $debs{$_} == 0, keys %debs;
+}
 
-    if (@gains) {
+if ($show_moved and $type ne 'deb') {
+    if (@debgains) {
 	my $msg = "Warning: these package names were in the second list but not in the first:";
 	print $msg, "\n", '-' x length $msg, "\n";
-	print join("\n",@gains), "\n\n";
+	print join("\n",@debgains), "\n\n";
     }
 
-    if (@losses) {
-	print "\n" if @gains;
+    if (@deblosses) {
+	print "\n" if @debgains;
 	my $msg = "Warning: these package names were in the first list but not in the second:";
 	print $msg, "\n", '-' x length $msg, "\n";
-	print join("\n",@losses), "\n\n";
-    }
-
-    # We run through all of the files in the first set of debs one by
-    # one, then pick up any which are new in the second set of debs
-    # We store any changed files in a hash of hashes %changes, where
-    # $changes{$from}{$to} is an array of files which have moved
-    # from package $from to package $to; $from or $to is '-' if
-    # the files have appeared or disappeared
-    my %changes;
-    foreach my $file (@D1) {
-	if (exists $files2{$file}) {
-	    next if $files1{$file} eq $files2{$file};
-	    # Ah, they're not the same.  We'll put a note in every
-	    # pair where the file is in the deb in the first set but not
-	    # in the deb in the second set
-	    my @firstdebs = split /:/, $files1{$file};
-	    my @seconddebs = split /:/, $files2{$file};
-	    foreach my $firstdeb (@firstdebs) {
-		foreach my $seconddeb (@seconddebs) {
-		    next if $firstdeb eq $seconddeb;
-		    push @{$changes{$firstdeb}{$seconddeb}}, $file;
-		}
-	    }
-	}
-	else {
-	    my @firstdebs = split /:/, $files1{$file};
-	    foreach my $firstdeb (@firstdebs) {
-		push @{$changes{$firstdeb}{'-'}}, $file;
-	    }
-	}
+	print join("\n",@deblosses), "\n\n";
     }
 
+    # We start by determining which files are in the first set of debs, the 
+    # second set of debs or both.
     my %files;
     grep $files{$_}--, @D1;
     grep $files{$_}++, @D2;
 
+    my @old = sort grep $files{$_} < 0, keys %files;
     my @new = sort grep $files{$_} > 0, keys %files;
+    my @same = sort grep $files{$_} == 0, keys %files;
+
+    # We store any changed files in a hash of hashes %changes, where
+    # $changes{$from}{$to} is an array of files which have moved
+    # from package $from to package $to; $from or $to is '-' if
+    # the files have appeared or disappeared
+
+    my %changes;
+
+    foreach my $file (@old) {
+	my @firstdebs = split /:/, $files1{$file};
+	foreach my $firstdeb (@firstdebs) {
+	    push @{$changes{$firstdeb}{'-'}}, $file;
+	}
+    }
 
     foreach my $file (@new) {
 	my @seconddebs = split /:/, $files2{$file};
@@ -471,6 +506,27 @@ if ($show_moved and $type ne 'deb') {
 	}
     }
 
+    foreach my $file (@same) {
+	# Are they identical?
+	next if $files1{$file} eq $files2{$file};
+
+	# Ah, they're not the same.  We'll put a note in every
+	# pair where the file is in the deb in the first set but not
+	# in the deb in the second set or vice versa
+	my %fdebs;
+	grep $fdebs{$_}--, split (/:/, $files1{$file});
+	grep $fdebs{$_}++, split (/:/, $files2{$file});
+	
+	my @firstdebs = sort grep $fdebs{$_} < 0, keys %fdebs;
+	my @seconddebs = sort grep $fdebs{$_} > 0, keys %fdebs;
+
+	foreach my $firstdeb (@firstdebs) {
+	    foreach my $seconddeb (@seconddebs) {
+		push @{$changes{$firstdeb}{$seconddeb}}, $file;
+	    }
+	}
+    }
+
     # This is not a very efficient way of doing things if there are
     # lots of debs involved, but since that is highly unlikely, it
     # shouldn't be much of an issue
@@ -481,6 +537,9 @@ if ($show_moved and $type ne 'deb') {
 	for my $deb2 ('-', sort keys %debs2) {
 	    next unless exists $changes{$deb1}{$deb2};
 	    my $msg;
+	    if (! $changed) {
+		print "[The following lists of changes regard files as different if they have\ndifferent names, permissions or owners.]\n\n";
+	    }
 	    if ($deb1 eq '-') {
 		$msg = "New files in second set of .debs, found in package $deb2";
 	    } elsif ($deb2 eq '-') {
@@ -506,8 +565,11 @@ if ($show_moved and $type ne 'deb') {
     my @losses = sort grep $files{$_} < 0, keys %files;
     my @gains = sort grep $files{$_} > 0, keys %files;
 
-    if (! $quiet && @losses == 0 && @gains == 0) {
-	print "File lists identical (after any substitutions)\n";
+    if (@losses == 0 && @gains == 0) {
+	print "File lists identical (after any substitutions)\n"
+	    unless $quiet;
+    } else {
+	print "[The following lists of changes regard files as different if they have\ndifferent names, permissions or owners.]\n\n";
     }
 
     if (@gains) {
@@ -539,56 +601,75 @@ if ($show_moved and $type ne 'deb') {
 }
 
 # We compare the control files (at least the dependency fields)
-# if we are examining precisely two .debs.
-exit $exit_status unless defined $singledeb[1] and defined $singledeb[2]
-    and $compare_control;
+my $dummyname = "---DUMMY---";
+if (defined $singledeb[1] and defined $singledeb[2]) {
+	@CommonDebs = ( $dummyname );
+	$DebPaths1{$dummyname} = $singledeb[1];
+	$DebPaths2{$dummyname} = $singledeb[2];
+}
+
+exit $exit_status unless (@CommonDebs > 0) and $compare_control;
 
 unless (system ("command -v wdiff >/dev/null 2>&1") == 0) {
     warn "Can't compare control files; wdiff package not installed\n";
     exit $exit_status;
 }
 
-mktmpdirs();
+for my $debname (@CommonDebs) {
+    no strict 'refs';
+    mktmpdirs();
 
-no strict 'refs';
-
-for my $i (1,2) {
-    if (system('dpkg-deb', '-e', "$singledeb[$i]", ${"dir$i"})) {
-	my $msg = "dpkg-deb -e $singledeb[$i] failed!";
-	system ("rm", "-rf", $dir1, $dir2);
-	fatal $msg;
+    for my $i (1,2) {
+	if (system('dpkg-deb', '-e', "${\"DebPaths$i\"}{$debname}", ${"dir$i"})) {
+	    my $msg = "dpkg-deb -e ${\"DebPaths$i\"}{$debname} failed!";
+	    system ("rm", "-rf", $dir1, $dir2);
+	    fatal $msg;
+	}
     }
-}
 
-use strict 'refs';
+    use strict 'refs';
 
-my $wdiff = `wdiff -n $wdiff_opt $dir1/control $dir2/control`;
-if ($? >> 8 == 0) {
-    if (! $quiet) {
-        print "\nNo differences were encountered in the control files\n";
-    }
-} elsif ($? >> 8 == 1) {
-    print "\n";
-    if ($wdiff_opt) {
-	# Don't try messing with control codes
-	my $msg = "The following is the wdiff output between the control files:";
-	print $msg, "\n", '-' x length $msg, "\n";
-	print $wdiff;
+    my @cf;
+    if ($controlfiles eq 'ALL') {
+	# only need to list one directory as we are only comparing control
+	# files in both packages
+	@cf = grep { ! /md5sums/ } map { basename($_); } glob("$dir1/*");
     } else {
-	my @output;
-	@output = split /\n/, $wdiff;
-	@output = grep /(\[-|\{\+)/, @output;
-	my $msg = "The following lines in the control files differ (wdiff output format):";
-	print $msg, "\n", '-' x length $msg, "\n";
-	print join("\n",@output), "\n";
+	@cf = split /,/, $controlfiles;
     }
-    $exit_status = 1;
-} else {
-    warn "wdiff failed (exit status " . ($? >> 8) .
-	(($? & 0x7f) ? " with signal " . ($? & 0x7f) : "") . ")\n";
+
+    foreach my $cf (@cf) {
+	next unless -f "$dir1/$cf" and -f "$dir2/$cf";
+	my $wdiff = `wdiff -n $wdiff_opt $dir1/$cf $dir2/$cf`;
+	my $usepkgname = $debname eq $dummyname ? "" : " of package $debname";
+	if ($? >> 8 == 0) {
+	    if (! $quiet) {
+		print "\nNo differences were encountered between the $cf files$usepkgname\n";
+	    }
+	} elsif ($? >> 8 == 1) {
+	    print "\n";
+	    if ($wdiff_opt) {
+		# Don't try messing with control codes
+		my $msg = ucfirst($cf) . " files$usepkgname: wdiff output";
+		print $msg, "\n", '-' x length $msg, "\n";
+		print $wdiff;
+	    } else {
+		my @output;
+		@output = split /\n/, $wdiff;
+		@output = grep /(\[-|\{\+)/, @output;
+		my $msg = ucfirst($cf) . " files$usepkgname: lines which differ (wdiff format)";
+		print $msg, "\n", '-' x length $msg, "\n";
+		print join("\n",@output), "\n";
+	    }
+	    $exit_status = 1;
+	} else {
+	    warn "wdiff failed (exit status " . ($? >> 8) .
+		(($? & 0x7f) ? " with signal " . ($? & 0x7f) : "") . ")\n";
+	}
+    }
+    # Clean up
+    system ("rm", "-rf", $dir1, $dir2);
 }
-# Clean up
-system ("rm", "-rf", $dir1, $dir2);
 
 exit $exit_status;
 
@@ -603,10 +684,8 @@ sub process_debc($$)
 
     # Format of dpkg-deb -c output:
     # permissions owner/group size date time name ['->' link destination]
-    # And remember the slink -> potato stuff
-    $data =~ s/^(\S+\s+){5}//mg;
-    $data =~ s,^\./,/,mg;
-    $data =~ s,^([^/]),/$1,mg;
+    $data =~ s/^(\S+)\s+(\S+)\s+(\S+\s+){3}/$1  $2   /mg;
+    $data =~ s, \./, /,mg;
     @filelist = grep ! m|^/$|, split /\n/, $data; # don't bother keeping '/'
 
     # Are we keeping directory names in our filelists?
@@ -621,13 +700,36 @@ sub process_debc($$)
 	    my $from  = $$move[1];
 	    my $to    = $$move[2];
 	    map { if ($regex) { eval "\$_ =~ s:^$from:$to:"; }
-	          else { $_ =~ s/^\Q$from\E/$to/; } } @filelist;
+		  else { $_ =~ s/^\Q$from\E/$to/; } } @filelist;
 	}
     }
 
     return \@filelist;
 }
 
+# This does the same for dpkg-deb -I
+sub process_debI($)
+{
+    my ($data) = @_;
+    my (@filelist);
+
+    # Format of dpkg-deb -c output:
+    # 2 (always?) header lines
+    #   nnnn bytes,    nnn lines   [*]  filename    [interpreter]
+    # Package: ...
+    # rest of control file
+
+    foreach (split /\n/, $data) {
+	last if /^Package:/;
+	next unless /^\s+\d+\s+bytes,\s+\d+\s+lines\s+(\*)?\s+([\-\w]+)/;
+	my $control = $2;
+	my $perms = ($1 ? "-rwxr-xr-x" : "-rw-r--r--");
+	push @filelist, "$perms  root/root   DEBIAN/$control";
+    }
+
+    return \@filelist;
+}
+
 sub mktmpdirs ()
 {
     no strict 'refs';
