@@ -63,6 +63,9 @@ my $modified_conf_msg;
 my @warnings;
 
 # Predeclare functions
+sub system_withecho(@);
+sub run_hook ($$);
+sub fileomitted (\@$);
 sub fatal($);
 
 sub usage
@@ -107,12 +110,33 @@ First usage method:
 Second usage method:
   $progname [<debuild options>] [<dpkg-buildpackage options>]
              [--lintian-opts <lintian options>] [--linda-opts <linda options>]
-    to run dpkg-buildpackage and then run lintian and/or linda on the resulting
-    .changes file.  Additional debuild option available in this case is:
-        --lintian                Run lintian (default)
-        --linda                  Run linda
-        --no-lintian             Do not run lintian
-        --no-linda               Do not run linda (default)
+    to run dpkg-buildpackage and then run lintian and/or linda on the
+    resulting .changes file.
+
+    Additional debuild options available in this case are:
+
+        --lintian           Run lintian (default)
+        --linda             Run linda
+        --no-lintian        Do not run lintian
+        --no-linda          Do not run linda (default)
+        --[no-]tgz-check    Do [not] check for an .orig.tar.gz before running
+                            dpkg-buildpackage if we have a Debian revision
+                            (Default: check) 
+
+        --dpkg-buildpackage-hook=HOOK
+        --clean-hook=HOOK
+        --dpkg-source-hook=HOOK
+        --build-hook=HOOK
+        --binary-hook=HOOK
+        --final-clean-hook=HOOK
+        --lintian-hook=HOOK
+        --signing-hook=HOOK
+        --post-dpkg-buildpackage-hook=HOOK
+                            These hooks run at the various stages of the
+                            dpkg-buildpackage run.  For details, see the
+                            debuild manpage.  They default to nothing, and
+                            can be reset to nothing with --foo-hook=''
+	--clear-hooks       Clear all hooks
 
     For available dpkg-buildpackage and lintian/linda options, see their
     respective manpages.
@@ -146,7 +170,7 @@ EOF
 # Set default values before we start
 my $preserve_env=0;
 my %save_vars;
-my $root_command='fakeroot';
+my $root_command='';
 my $run_lintian=1;
 my $run_linda=0;
 my $lintian_exists=0;
@@ -156,10 +180,15 @@ my @lintian_extra_opts=();
 my @lintian_opts=();
 my @linda_extra_opts=();
 my @linda_opts;
-my $run_builddeps=1;
+my $checkbuilddep=1;
 my $check_dirname_level = 1;
 my $check_dirname_regex = 'PACKAGE(-.*)?';
 my $logging=0;
+my $tgz_check=1;
+my @hooks = (qw(dpkg-buildpackage clean dpkg-source build binary final-clean
+		lintian signing post-dpkg-buildpackage));
+my %hook;
+$hook{@hooks} = ('') x @hooks;
 
 
 # First handle private options from cvs-debuild
@@ -213,6 +242,16 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
 		       'DEBUILD_LINTIAN' => 'yes',
 		       'DEBUILD_LINDA' => 'no',
 		       'DEBUILD_ROOTCMD' => 'fakeroot',
+		       'DEBUILD_TGZ_CHECK' => 'yes',
+		       'DEBUILD_DPKG_BUILDPACKAGE_HOOK' => '',
+		       'DEBUILD_CLEAN_HOOK' => '',
+		       'DEBUILD_DPKG_SOURCE_HOOK' => '',
+		       'DEBUILD_BUILD_HOOK' => '',
+		       'DEBUILD_BINARY_HOOK' => '',
+		       'DEBUILD_FINAL_CLEAN_HOOK' => '',
+		       'DEBUILD_LINTIAN_HOOK' => '',
+		       'DEBUILD_SIGNING_HOOK' => '',
+		       'DEBUILD_POST_DPKG_BUILDPACKAGE_HOOK' => '',
 		       'DEVSCRIPTS_CHECK_DIRNAME_LEVEL' => 1,
 		       'DEVSCRIPTS_CHECK_DIRNAME_REGEX' => 'PACKAGE(-.*)?',
 		       );
@@ -258,6 +297,8 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
 	or $config_vars{'DEBUILD_LINDA'}='no';
     $config_vars{'DEVSCRIPTS_CHECK_DIRNAME_LEVEL'} =~ /^[012]$/
 	or $config_vars{'DEVSCRIPTS_CHECK_DIRNAME_LEVEL'}=1;
+    $config_vars{'DEBUILD_TGZ_CHECK'} =~ /^(yes|no)$/
+	or $config_vars{'DEBUILD_TGZ_CHECK'}='yes';
 
     foreach my $var (sort keys %config_vars) {
 	if ($config_vars{$var} ne $config_default{$var}) {
@@ -275,8 +316,14 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
     $run_lintian = $config_vars{'DEBUILD_LINTIAN'} eq 'no' ? 0 : 1;
     $run_linda = $config_vars{'DEBUILD_LINDA'} eq 'yes' ? 1 : 0;
     $root_command = $config_vars{'DEBUILD_ROOTCMD'};
+    $tgz_check = $config_vars{'DEBUILD_TGZ_CHECK'} eq 'yes' ? 1 : 0;
     $check_dirname_level = $config_vars{'DEVSCRIPTS_CHECK_DIRNAME_LEVEL'};
     $check_dirname_regex = $config_vars{'DEVSCRIPTS_CHECK_DIRNAME_REGEX'};
+    for my $hookname (@hooks) {
+	my $config_name = uc "debuild_${hookname}_hook";
+	$config_name =~ tr/-/_/;
+	$hook{$hookname} = $config_vars{$config_name};
+    }
 
     # Now parse the opts lists
     if (shift @othervars ne ">>> $dpkg_opts_var BEGIN <<<") {
@@ -337,12 +384,14 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
 # We first check @dpkg_extra_opts for options which may affect us;
 # these were set in a configuration file, so they have lower
 # precedence than command line settings.  The options we care about
-# at this stage are: -r, -d and -D
+# at this stage are: -r and those which affect the checkbuilddep setting
 
 foreach (@dpkg_extra_opts) {
     /^-r(.*)$/ and $root_command=$1, next;
-    $_ eq '-d' and $run_builddeps=0, next;
-    $_ eq '-D' and $run_builddeps=1, next;
+    $_ eq '-d' and $checkbuilddep=0, next;
+    $_ eq '-D' and $checkbuilddep=1, next;
+    /^-a(.*)/ and $_ ne '-ap' and $checkbuilddep=0, next;
+    $_ eq '-S' and $checkbuilddep=0, next;
 }
 
 # Check @ARGV for debuild options.
@@ -449,6 +498,8 @@ my @preserve_vars = qw(TERM HOME LOGNAME PGPPATH GNUPGHOME GPG_AGENT_INFO
 	    $root_command=$opt;
 	    next;
 	}
+	$arg eq '--tgz-check' and $tgz_check=1, next;
+	$arg =~ /^--no-?tgz-check$/ and $tgz_check=0, next;
 	$arg =~ /^-r(.*)/ and $root_command=$1, next;
 	if ($arg eq '--ignore-dirname') {
 	    fatal "--ignore-dirname has been replaced by --check-dirname-level and\n--check-dirname-regex; run $progname --help for more details";
@@ -481,8 +532,36 @@ my @preserve_vars = qw(TERM HOME LOGNAME PGPPATH GNUPGHOME GPG_AGENT_INFO
 	if ($arg =~ /^--no-?conf$/) {
 	    fatal "$arg is only acceptable as the first command-line option!";
 	}
-	$arg eq '-d' and $run_builddeps=0, next;
-	$arg eq '-D' and $run_builddeps=1, next;
+	$arg eq '-d' and $checkbuilddep=0, next;
+	$arg eq '-D' and $checkbuilddep=1, next;
+
+	# hooks...
+	if ($arg =~ /^--(.*)-hook$/) {
+	    my $argkey = $1;
+	    unless (exists $hook{$argkey}) {
+		fatal "unknown hook $arg,\nrun $progname --help for usage information";
+	    }
+	    unless (defined ($opt = shift)) {
+		fatal "$arg requires an argument,\nrun $progname --help for usage information";
+	    }
+	    $hook{$argkey} = $opt;
+	    next;
+	}
+
+	if ($arg =~ /^--(.*?)-hook=(.*)/) {
+	    my $argkey = $1;
+	    my $opt = $2;
+
+	    unless (exists $hook{$argkey}) {
+		fatal "unknown hook option $arg,\nrun $progname --help for usage information";
+	    }
+
+	    $hook{$argkey} = $opt;
+	    next;
+	}
+
+	if ($arg eq '--clear-hooks') { $hook{@hooks} = ('') x @hooks; next; }
+
 	# Not a debuild option, so give up.
 	unshift @ARGV, $arg;
 	last;
@@ -609,7 +688,29 @@ else {
 }
 
 if ($command_version eq 'dpkg') {
-    # We're going to run dpkg-buildpackage and possibly lintian/linda.
+    # We're going to emulate dpkg-buildpackage and possibly lintian/linda.
+    # This will allow us to run hooks.
+    # However, if dpkg-cross is installed (as evidenced by the presence
+    # of /usr/bin/dpkg-cross), then we call the "real" dpkg-buildpackage,
+    # which is actually dpkg-cross's version.  We lose the facility for
+    # hooks in this case, but we're not going to emulate dpkg-cross as well!
+
+    my $dpkg_cross = (-x '/usr/bin/dpkg-cross' ? 1 : 0);
+    if ($dpkg_cross) {
+	# check hooks
+	my @skip_hooks = ();
+	for my $hookname (qw(clean dpkg-source build binary dpkg-genchanges
+			     final-clean)) {
+	    if ($hook{$hookname}) { push @skip_hooks, $hookname; }
+	}
+	if (@skip_hooks) {
+	    warn "$progname: dpkg-cross appears to be present on your system, and you have\nset up some hooks which will not be run (" .
+		join(", ", @skip_hooks) . ");\ndo you wish to continue? (y/n) ";
+	    my $ans = <STDIN>;
+	    exit 1 unless $ans =~ /^y/i;
+	}
+    }
+
     # Our first task is to parse the command line options.
 
     # And before we get too excited, does lintian/linda even exist?
@@ -622,53 +723,128 @@ if ($command_version eq 'dpkg') {
 	    and $linda_exists=1;
     }
 
-    my $Lopts=0;
+    # dpkg-buildpackage variables explicitly initialised in dpkg-buildpackage
+    my $signsource=1;
+    my $signchanges=1;
+    my $cleansource=0;
+    my $binarytarget='binary';
+    my $sourcestyle='';
+    my $since='';
+    my $maint='';
+    my $desc='';
+    my $noclean=0;
+    my $usepause=0;
+    my $warnable_error=0;  # OK, dpkg-buildpackage defines this but doesn't
+                           # use it.  We'll keep it around just in case it
+                           # does one day...
+    my @passopts=();
+
+    # extra dpkg-buildpackage variables not initialised there
+    my $diffignore='';
+    my @tarignore=();
     my $sourceonly='';
     my $binaryonly='';
     my $targetarch='';
-    my $gnutarget='';
-    my $signchanges=1;
-    my $signdsc=1;
-    my @dpkg_opts = qw(-us -uc);
-    my @debsign_opts = ();
+    my $targetgnusystem='';
+    my $changedby='';
 
+    # and one for us
+    my @debsign_opts = ();
+    # and one for dpkg-cross if needed
+    my @dpkg_opts = qw(-us -uc);
+
+    # Parse dpkg-buildpackage options
     # First process @dpkg_extra_opts from above
 
     foreach (@dpkg_extra_opts) {
-	/^-r(.*)/ and next;  # already been processed
+	$_ eq '-h' and
+	    warn "You have a -h option in your configuration file!  Ignoring.\n", next;
+	/^-r/ and next;  # already been processed
+	/^-p/ and push(@debsign_opts, $_), next;  # Key selection options
+	/^-k/ and push(@debsign_opts, $_), next;  # Ditto
 	/^-[dD]$/ and next;  # already been processed
-	/^-a(.*)/ and $targetarch=$1;       # Explained below
-	/^-t(.*)/ and $_ ne '-tc' and $gnutarget=$1;    # Ditto
-	$_ eq '-S' and $sourceonly=$_;       # Explained below
-	/^-[mek]/ and push @debsign_opts, $_;    # Key selection options
-	/^-s(pgp|gpg)$/ and push @debsign_opts, $_;  # Ditto
-	/^-p/ and push @debsign_opts, $_;  # Ditto
-	$_ eq '-us' and $signdsc=0, next;
+	/^-s(pgp|gpg)$/ and push(@debsign_opts, $_), next;  # Key selection
+	$_ eq '-us' and $signsource=0, next;
 	$_ eq '-uc' and $signchanges=0, next;
-	/^-[Bb]$/ and $binaryonly=$_;
-	push @dpkg_opts, $_;
+	$_ eq '-ap' and $usepause=1, next;
+	/^-a(.*)/ and $targetarch=$1, push(@dpkg_opts, $_), next;
+	    # Explained below; no implied -d here, as already done
+	/^-s[iad]$/ and $sourcestyle=$_, push(@dpkg_opts, $_), next;
+	/^-i/ and $diffignore=$_, push(@dpkg_opts, $_), next;
+	/^-I/ and push(@tarignore, $_), push(@dpkg_opts, $_), next;
+	$_ eq '-tc' and $cleansource=1, push(@dpkg_opts, $_), next;
+	/^-t(.*)/ and $targetgnusystem=$1, push(@dpkg_opts, $_), next; # Ditto	
+	$_ eq '-nc' and $noclean=1, $binaryonly ||= '-b', push(@dpkg_opts, $_),
+	    next;
+	$_ eq '-b' and $binaryonly=$_, push(@dpkg_opts, $_), next;
+	$_ eq '-B' and $binaryonly=$_, $binarytarget='binary-arch',
+	    push(@dpkg_opts, $_), next;
+	$_ eq '-S' and $sourceonly=$_, push(@dpkg_opts, $_), next;
+	    # Explained below, no implied -d
+	/^-v(.*)/ and $since=$1, push(@dpkg_opts, $_), next;
+	/^-m(.*)/ and $maint=$1, push(@debsign_opts, $_), push(@dpkg_opts, $_),
+	    next;
+	/^-e(.*)/ and $changedby=$1, push(@debsign_opts, $_),
+	    push(@dpkg_opts, $_), next;
+	/^-C(.*)/ and $desc=$1, push(@dpkg_opts, $_), next;
+	$_ eq '-W' and $warnable_error=1, push(@passopts, $_),
+	    push(@dpkg_opts, $_), next;
+	$_ eq '-E' and $warnable_error=0, push(@passopts, $_),
+	    push(@dpkg_opts, $_), next;
+	# dpkg-cross specific option
+	if (/^-M/ and $dpkg_cross) { push(@dpkg_opts, $_), next; }
+	fatal "unknown dpkg-buildpackage option in configuration file: $_";
     }
 
     while ($_=shift) {
+	$_ eq '-h' and usage(), exit 0;
 	/^-r(.*)/ and $root_command=$1, next;
-	$_ eq '-d' and $run_builddeps=0, next;
-	$_ eq '-D' and $run_builddeps=1, next;
-	/^-a(.*)/ and $targetarch=$1;       # Explained below
-	/^-t(.*)/ and $_ ne '-tc' and $gnutarget=$1;    # Ditto
-	$_ eq '-S' and $sourceonly=$_;       # Explained below
-	/^-[mek]/ and push @debsign_opts, $_;     # Key selection options
-	/^-s(pgp|gpg)$/ and push @debsign_opts, $_;  # Ditto
-	/^-p/ and push @debsign_opts, $_;  # Ditto
-	$_ eq '-us' and $signdsc=0, next;
+	/^-p/ and push(@debsign_opts, $_), next;  # Key selection options
+	/^-k/ and push(@debsign_opts, $_), next;  # Ditto
+	$_ eq '-d' and $checkbuilddep=0, next;
+	$_ eq '-D' and $checkbuilddep=1, next;
+	/^-s(pgp|gpg)$/ and push(@debsign_opts, $_), next;  # Key selection
+	$_ eq '-us' and $signsource=0, next;
 	$_ eq '-uc' and $signchanges=0, next;
-	/^-[Bb]$/ and $binaryonly=$_;
+	$_ eq '-ap' and $usepause=1, next;
+	/^-a(.*)/ and $targetarch=$1, $checkbuilddep=0, push(@dpkg_opts, $_),
+	    next;
+	/^-s[iad]$/ and $sourcestyle=$_, push(@dpkg_opts, $_), next;
+	/^-i/ and $diffignore=$_, push(@dpkg_opts, $_), next;
+	/^-I/ and push(@tarignore, $_), push(@dpkg_opts, $_), next;
+	$_ eq '-tc' and $cleansource=1, push(@dpkg_opts, $_), next;
+	/^-t(.*)/ and $targetgnusystem=$1, $checkbuilddep=0, next;
+	$_ eq '-nc' and $noclean=1, $binaryonly ||= '-b', push(@dpkg_opts, $_),
+	    next;
+	$_ eq '-b' and $binaryonly=$_, push(@dpkg_opts, $_), next;
+	$_ eq '-B' and $binaryonly=$_, $binarytarget='binary-arch',
+	    push(@dpkg_opts, $_), next;
+	$_ eq '-S' and $sourceonly=$_, $checkbuilddep=0, push(@dpkg_opts, $_),
+	    next;
+	/^-v(.*)/ and $since=$1, push(@dpkg_opts, $_), next;
+	/^-m(.*)/ and $maint=$1, push(@debsign_opts, $_), push(@dpkg_opts, $_),
+	    next;
+	/^-e(.*)/ and $changedby=$1, push(@debsign_opts, $_),
+	    push(@dpkg_opts, $_), next;
+	/^-C(.*)/ and $desc=$1, push(@dpkg_opts, $_), next;
+	$_ eq '-W' and $warnable_error=1, push(@passopts, $_),
+	    push(@dpkg_opts, $_), next;
+	$_ eq '-E' and $warnable_error=0, push(@passopts, $_),
+	    push(@dpkg_opts, $_), next;
+	# dpkg-cross specific option
+	if (/^-M/ and $dpkg_cross) { push(@dpkg_opts, $_), next; }
+
+	# these non-dpkg-buildpackage options make us stop
 	if ($_ eq '-L' or $_ eq '--lintian' or /^--(lintian|linda)-opts$/) {
 	    unshift @ARGV, $_;
 	    last;
 	}
-	push @dpkg_opts, $_;
+	fatal "unknown dpkg-buildpackage/debuild option: $_";
+    }
+
+    if ($sourceonly and $binaryonly) {
+	fatal "cannot combine dpkg-buildpackage options $sourceonly and $binaryonly";
     }
-    unshift @dpkg_opts, ($run_builddeps ? "-D" : "-d");
 
     # Pick up lintian/linda options if necessary
     if (($run_lintian || $run_linda) && @ARGV) {
@@ -711,47 +887,60 @@ if ($command_version eq 'dpkg') {
     }
 
     if ($< != 0) {
-	if ($root_command) {
-	    # Only fakeroot is a default, so that's the only one we'll
-	    # check for
-	    if ($root_command eq 'fakeroot') {
-		system('fakeroot true 2>/dev/null');
-		if ($? >> 8 != 0) {
-		    fatal "problem running fakeroot: either install the fakeroot package,\nuse a -r option to select another root command program to use or\nrun me as root!";
-		}
+	$root_command ||= 'fakeroot';
+	# Only fakeroot is a default, so that's the only one we'll
+	# check for
+	if ($root_command eq 'fakeroot') {
+	    system('fakeroot true 2>/dev/null');
+	    if ($? >> 8 != 0) {
+		fatal "problem running fakeroot: either install the fakeroot package,\nuse a -r option to select another root command program to use or\nrun me as root!";
 	    }
-	    unshift @dpkg_opts, "-r$root_command";
-	} else {
-	    fatal "need a --rootcmd or -r option to run!";
 	}
     }
 
-    if ($signchanges==1 and $signdsc==0) {
+    if ($signchanges==1 and $signsource==0) {
 	push @warnings,
 	    "I will sign the .dsc file anyway as a signed .changes file was requested\n";
+	$signsource=1;  # may not be strictly necessary, but for clarity!
     }
 
-    # We need to figure out what the changes file will be called,
-    # so we copy some code from dpkg-buildpackage for this purpose.
-    # Note that dpkg-buildpackage looks at any -a... and -t... parameters
-    # it is given to determine the architecture, so we need to do the
-    # same to determine the .changes filename.
+    # Next dpkg-buildpackage steps:
+    # mustsetvar package/version have been done above; we've called the
+    # results $pkg and $version
+    # mustsetvar maintainer is only needed for signing, so we leave that
+    # to debsign or dpkg-sig
+    # Call to dpkg-architecture to set DEB_{BUILD,HOST}_* environment
+    # variables
+    foreach (split /\n/, `dpkg-architecture -a${targetarch} -t${targetgnusystem} -f`) {
+	/^(.*)=(.*)$/ and $ENV{$1} = $2;
+    }
 
-    # The following is based on dpkg-buildpackage
-    my ($sversion, $dsc, $changes, $build);
+    # We need to do the arch, sversion, pv, pva stuff to figure out
+    # what the changes file will be called,
+    my ($sversion, $uversion, $dsc, $changes, $build);
     my $arch;
     if ($sourceonly) {
-	$arch = 'source' ;
+	$arch = 'source';
     } else {
-	$arch=`dpkg-architecture -a${targetarch} -t${gnutarget} -qDEB_HOST_ARCH`;
+	$arch=`dpkg-architecture -a${targetarch} -t${targetgnusystem} -qDEB_HOST_ARCH`;
 	chomp $arch;
-	fatal "couldn't determine architecture!?" if ! $arch;
+	fatal "couldn't determine host architecture!?" if ! $arch;
     }
 
     ($sversion=$version) =~ s/^\d+://;
-    $dsc="${pkg}_$sversion.dsc";
-    $changes="${pkg}_${sversion}_${arch}.changes";
+    ($uversion=$sversion) =~ s/-[a-z0-9+\.]+$//i;
+    $dsc="${pkg}_${sversion}.dsc";
+    my $origtgz="${pkg}_${uversion}.orig.tar.gz";
+    if ($tgz_check and $uversion ne $sversion and ! -f "../$origtgz") {
+	print STDERR "This package has a Debian revision number but there does not seem to be\nan appropriate original tar file in the parent directory;\n(expected $origtgz)\ncontinue anyway? (y/n) ";
+	my $ans = <STDIN>;
+	exit 1 unless $ans =~ /^y/i;
+    }
+
+    # We'll need to be a bit cleverer to determine the changes file name;
+    # see below
     $build="${pkg}_${sversion}_${arch}.build";
+    $changes="${pkg}_${sversion}_${arch}.changes";
     open BUILD, "| tee ../$build" or fatal "couldn't open pipe to tee: $!";
     $logging=1;
     close STDOUT;
@@ -759,13 +948,179 @@ if ($command_version eq 'dpkg') {
     open STDOUT, ">&BUILD" or fatal "can't reopen stdout: $!";
     open STDERR, ">&BUILD" or fatal "can't reopen stderr: $!";
 
-    # So now we can run dpkg-buildpackage and lintian/linda...
+    run_hook('dpkg-buildpackage', 1);
 
-    # print STDERR "Running dpkg-buildpackage @dpkg_opts\n";
-    system('dpkg-buildpackage', @dpkg_opts) == 0
-	or fatal "dpkg-buildpackage failed!";
-    chdir '..' or fatal "can't chdir: $!";
+    if ($dpkg_cross) {
+	unshift @dpkg_opts, ($checkbuilddep ? "-D" : "-d");
+	unshift @dpkg_opts, "-r$root_command" if $root_command;
+	system_withecho('dpkg-buildpackage', @dpkg_opts);
 
+	chdir '..' or fatal "can't chdir: $!";
+	# We're using dpkg-cross, we could now have foo_1.2_i386+sparc.changes
+	# so can't just set $changes = "${pkg}_${sversion}_${arch}.changes"
+	my @changes = glob("${pkg}_${sversion}_*.changes");
+	if (@changes == 0) {
+	    fatal "couldn't find a .changes file!";
+	} elsif (@changes == 1) {
+	    $changes = $changes[0];
+	} else {
+	    # put newest first
+	    @changes = sort { -M $a <=> -M $b } @changes;
+	    $changes = $changes[0];
+	}
+    } else {
+	# Not using dpkg-cross, so we emulate dpkg-buildpackage ourselves
+	# We emulate the version found in dpkg-buildpackage-snapshot in
+	# the source package
+
+	# First dpkg-buildpackage action: run dpkg-checkbuilddeps
+	if ($checkbuilddep) {
+	    if ($binarytarget eq 'binary-arch') {
+		system('dpkg-checkbuilddeps -B');
+	    } else {
+		system('dpkg-checkbuilddeps');
+	    }
+	    if ($?>>8) {
+		# Horrible non-Perlish formatting here, but emacs formatting
+		# dies miserably if this paragraph is done as a here-document.
+		# And even the documented hack doesn't work here :(
+		fatal "You do not appear to have all build dependencies properly met, aborting.\n" .
+		    "(Use -d flag to override.)\n" .
+		    "If you have the pbuilder package installed you can run\n" .
+		    "/usr/lib/pbuilder/pbuilder-satisfydepends as root to install the\n" .
+		    "required packages, or you can do it manually using dpkg or apt using\n" .
+		    "the error messages just above this message.\n";
+	    }
+	}
+
+	run_hook('clean', ! $noclean);
+
+	# Next dpkg-buildpackage action: clean
+	unless ($noclean) {
+	    if ($< == 0) {
+		system_withecho('debian/rules', 'clean');
+	    } else {
+		system_withecho($root_command, 'debian/rules', 'clean');
+	    }
+	}
+
+	run_hook('dpkg-source', ! $binaryonly);
+
+	# Next dpkg-buildpackage action: dpkg-source
+	if (! $binaryonly) {
+	    my $dirn = basename(cwd());
+	    my @cmd = (qw(dpkg-source));
+	    push @cmd, @passopts;
+	    push @cmd, $diffignore if $diffignore;
+	    push @cmd, @tarignore;
+	    push @cmd, "-b", $dirn;
+	    chdir '..' or fatal "can't chdir ..: $!";
+	    system_withecho(@cmd);	
+	    chdir $dirn or fatal "can't chdir $dirn: $!";
+	}
+
+	run_hook('build', ! $sourceonly);
+
+	# Next dpkg-buildpackage action: build and binary targets
+	if (! $sourceonly) {
+	    system_withecho('debian/rules', 'build');
+	
+	    run_hook('binary', 1);
+
+	    if ($< == 0) {
+		system_withecho('debian/rules', $binarytarget);
+	    } else {
+		system_withecho($root_command, 'debian/rules', $binarytarget);
+	    }
+	} elsif ($hook{'binary'}) {
+	    push @warnings, "$progname: not running binary hook '$hook{'binary'}' as -S option used\n";
+	}
+
+	# We defer the signing the .dsc file until after dpkg-genchanges has
+	# been run
+
+	run_hook('dpkg-genchanges', 1);
+
+	# Because of our messing around with STDOUT and wanting to pass
+	# arguments safely to dpkg-genchanges means that we're gonna have to
+	# do it manually :(
+	my @cmd = ('dpkg-genchanges');
+	foreach ($binaryonly, $sourceonly, $sourcestyle) {
+	    push @cmd, $_ if $_;
+	}
+	push @cmd, "-m$maint" if $maint;
+	push @cmd, "-e$changedby" if $changedby;
+	push @cmd, "-v$since" if $since;
+	push @cmd, "-C$desc" if $desc;
+	print STDERR " ", join(" ", @cmd), "\n";
+	
+	open GENCHANGES, "-|", @cmd or fatal "can't exec dpkg-genchanges: $!";
+	my @changefilecontents;
+	@changefilecontents = <GENCHANGES>;
+	close GENCHANGES
+	    or warn "$progname: dpkg-genchanges failed!\n", exit ($?>>8);
+	open CHANGES, "> ../$changes"
+	    or fatal "can't open ../$changes for writing: $!";
+	print CHANGES @changefilecontents;
+	close CHANGES
+	    or fatal "problem writing to ../$changes: $!";
+
+	run_hook('final-clean', $cleansource);
+
+	# Final dpkg-buildpackage action: clean target again
+	if ($cleansource) {
+	    if ($< == 0) {
+		system_withecho('debian/rules', 'clean');
+	    } else {
+		system_withecho($root_command, 'debian/rules', 'clean');
+	    }
+	}
+
+
+	# identify the files listed in $changes; this will be used for the
+	# emulation of the dpkg-buildpackage fileomitted() function
+
+	my @files;
+	my $infiles=0;
+	foreach (@changefilecontents) {
+	    /^Files:/ and $infiles=1, next;
+	    next unless $infiles;
+	    last if /^[^ ]/; # no need to go further
+	    # so we're looking at a filename with lots of info before it
+	    / (\S+)$/ and push @files, $1;
+	}
+
+	my $srcmsg;
+
+	if (fileomitted @files, '\.deb') {
+	    # source only upload
+	    if (fileomitted @files, '\.diff\.gz') {
+		$srcmsg='source only upload: Debian-native package';
+	    } elsif (fileomitted @files, '\.orig\.tar\.gz') {
+		$srcmsg='source only, diff-only upload (original source NOT included)';
+	    } else {
+		$srcmsg='source only upload (original source is included)';
+	    }
+	} else {
+	    if (fileomitted @files, '\.dsc') {
+		$srcmsg='binary only upload (no source included)'
+	    } elsif (fileomitted @files, '\.diff\.gz') {
+		$srcmsg='full upload; Debian-native package (full source is included)';
+	    } elsif (fileomitted @files, '\.orig\.tar\.gz') {
+		$srcmsg='binary and diff upload (original source NOT included)';
+	    } else {
+		$srcmsg='full upload (original source is included)';
+	    }
+	}
+
+	print "dpkg-buildpackage (debuild emulation): $srcmsg\n";
+
+	chdir '..' or fatal "can't chdir: $!";
+    } # end of debuild dpkg-buildpackage emulation
+
+    run_hook('lintian', (($run_lintian && $lintian_exists) ||
+			 ($run_linda && $linda_exists)) );
+    
     if ($run_lintian && $lintian_exists) {
 	$<=$>=$uid;  # Give up on root privileges if we can
 	$(=$)=$gid;
@@ -783,17 +1138,27 @@ if ($command_version eq 'dpkg') {
 	print "Finished running linda.\n";
     }
 
+    # They've insisted.  Who knows why?!
+    if (($signchanges or $signsource) and $usepause) {
+	print "Press the return key to start signing process\n";
+	<STDIN>;
+    }
+
+    run_hook('signing', ($signchanges || (! $sourceonly and $signsource)) );
+
     if ($signchanges) {
 	print "Now signing changes and any dsc files...\n";
 	system('debsign', @debsign_opts, $changes) == 0
 	    or fatal "running debsign failed";
     }
-    elsif (! $sourceonly and $signdsc) {
+    elsif (! $sourceonly and $signsource) {
 	print "Now signing dsc file...\n";
 	system('debsign', @debsign_opts, $dsc) == 0
 	    or fatal "running debsign failed";
     }
 
+    run_hook('post-dpkg-buildpackage', 1);
+
     # Any warnings?
     if (@warnings) {
 	# Don't know why we need this, but seems that we do, otherwise,
@@ -801,7 +1166,7 @@ if ($command_version eq 'dpkg') {
 	IO::Handle::flush(\*STDOUT);
 
 	my $warns = @warnings > 1 ? "S" : "";
-	warn "\nWARNING$warns generated by debuild:\n" .
+	warn "\nWARNING$warns generated by $progname:\n" .
 	    join("\n", @warnings) . "\n";
     }
     # close the logging process
@@ -814,21 +1179,20 @@ if ($command_version eq 'dpkg') {
 }
 else {
     # Running debian/rules.  Do dpkg-checkbuilddeps first
-    if ($run_builddeps) {
+    if ($checkbuilddep) {
 	if ($ARGV[0] eq 'binary-arch') {
 	    system('dpkg-checkbuilddeps -B');
 	} else {
 	    system('dpkg-checkbuilddeps');
 	}
 	if ($?>>8) {
-	    warn <<EOT;
+	    fatal <<"EOT";
 You do not appear to have all build dependencies properly met.
 If you have the pbuilder package installed, you can run
 /usr/lib/pbuilder/pbuilder-satisfydepends as root to install the
 required packages, or you can do it manually using dpkg or apt using
 the error messages just above this message.
 EOT
-	    exit ($?>>8);
 	}
     }
 
@@ -854,6 +1218,38 @@ EOT
 
 ###### Subroutines
 
+sub system_withecho(@) {
+    print STDERR " ", join(" ", @_), "\n";
+    system(@_);
+    if ($?>>8) {
+	fatal "@_ failed";
+    }
+}
+
+sub fileomitted (\@$) {
+    my ($files, $pat) = @_;
+    return (scalar(grep { /$pat$/ } @$files) == 0);
+}
+
+sub run_hook ($$) {
+    my ($hook, $act) = @_;
+    return unless $hook{$hook};
+
+    print STDERR " Running $hook-hook\n";
+    my $hookcmd = $hook{$hook};
+    $act = $act ? 1 : 0;
+    my %per=("%"=>"%", "p"=>$pkg, "v"=>$version, "a"=>$act);
+    $hookcmd =~ s/\%(.)/exists $per{$1} ? $per{$1} :
+	(warn ("Unrecognised \% substitution in hook: \%$1\n"), "\%$1")/eg;
+
+    system_withecho($hookcmd);
+
+    if ($?>>8) {
+	warn "$progname: $hook-hook failed\n";
+	exit ($?>>8);
+    }
+}
+
 sub fatal($) {
     my ($pack,$file,$line);
     ($pack,$file,$line) = caller();
