@@ -6,8 +6,7 @@
 # Originally written by Christoph Lameter <clameter@debian.org> (I believe)
 # Modified by Julian Gilbey <jdg@debian.org>
 # HTTP support added by Piotr Roszatycki <dexter@debian.org>
-# Copyright 1999, Julian Gilbey
-# Rewritten in Perl, Copyright 2002, Julian Gilbey
+# Rewritten in Perl, Copyright 2002-2006, Julian Gilbey
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -122,7 +121,7 @@ EOF
 sub version {
     print <<"EOF";
 This is $progname, from the Debian devscripts package, version ###VERSION###
-This code is copyright 1999 by Julian Gilbey, all rights reserved.
+This code is copyright 1999-2006 by Julian Gilbey, all rights reserved.
 Original code by Christoph Lameter.
 This program comes with ABSOLUTELY NO WARRANTY.
 You are free to redistribute this code under the terms of the
@@ -189,7 +188,7 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
 	or $config_vars{'USCAN_SYMLINK'}='yes';
     $config_vars{'USCAN_SYMLINK'}='symlink'
 	if $config_vars{'USCAN_SYMLINK'} eq 'yes' or
-	    $config_vars{'USCAN_SYMLINK'} eq 'symlinks';
+	    $config_vars{'USCAN_SYMLINK'} =~ /^symlinks?$/;
     $config_vars{'USCAN_VERBOSE'} =~ /^(yes|no)$/
 	or $config_vars{'USCAN_VERBOSE'}='no';
     $config_vars{'USCAN_DEHS_OUTPUT'} =~ /^(yes|no)$/
@@ -389,20 +388,14 @@ print "-- Scanning for watchfiles in @ARGV\n" if $verbose;
 # correctly, which makes this code a little messier than it would be
 # otherwise.
 my @dirs;
-my $pid = open FIND, '-|';
-if (! defined $pid) {
-    die "$progname: couldn't fork: $!\n";
-}
-if ($pid) {
-    while (<FIND>) {
-	chomp;
-	push @dirs, $_;
-    }
-    close FIND;
-} else {
-    exec 'find', @ARGV, qw(-follow -type d -name debian -print);
-    die "$progname: couldn't exec find: $!\n";
+open FIND, '-|', 'find', @ARGV, qw(-follow -type d -name debian -print)
+    or die "$progname: couldn't exec find: $!\n";
+
+while (<FIND>) {
+    chomp;
+    push @dirs, $_;
 }
+close FIND;
 
 die "$progname: No debian directories found\n" unless @dirs;
 
@@ -571,13 +564,36 @@ exit $found ? 0 : 1;
 # concatenated before mangling is performed, and that mangling will
 # only be performed on the basename version number, not any path version
 # numbers.)
-# opts=uversionmangle=s/^/0.0/ \
+# opts=uversionmangle=s/^/0.0./ \
 #   ftp://ftp.ibiblio.org/pub/Linux/ALPHA/wine/development/Wine-(.*)\.tar\.gz
 # 
 # Similarly, the upstream part of the Debian version number can be
 # mangled:
 # opts=dversionmangle=s/\.dfsg\.\d+$// \
 #   http://some.site.org/some/path/foobar-(.*)\.tar\.gz
+# 
+# The versionmangle=... option is a shorthand for saying uversionmangle=...
+# and dversionmangle=... and applies to both upstream and Debian versions.
+# 
+# The option filenamemangle can be used to mangle the name under which
+# the downloaded file will be saved:
+#   href="http://foo.bar.org/download/?path=&amp;download=foo-0.1.1.tar.gz"
+# could be handled as:
+# opts=filenamemangle=s/.*=(.*)/$1/ \
+#     http://foo.bar.org/download/\?path=&amp;download=foo-(.*)\.tar\.gz
+# and
+#   href="http://foo.bar.org/download/?path=&amp;download_version=0.1.1"
+# as:
+# opts=filenamemangle=s/.*=(.*)/foo-$1\.tar\.gz/ \
+#    http://foo.bar.org/download/\?path=&amp;download_version=(.*)
+# 
+# The option downloadurlmangle can be used to mangle the URL of the file
+# to download.  This can only be used with http:// URLs.  This may be
+# necessary if the link given on the webpage needs to be transformed in
+# some way into one which will work automatically, for example:
+# opts=downloadurlmangle=s/prdownload/download/ \
+#   http://developer.berlios.de/project/showfiles.php?group_id=2051 \
+#   http://prdownload.berlios.de/softdevice/vdr-softdevice-(.*).tgz
 
 
 sub process_watchline ($$$$$$)
@@ -596,7 +612,7 @@ sub process_watchline ($$$$$$)
     %dehs_tags = ('package' => $pkg);
 
     if ($watch_version == 1) {
-	($site, $dir, $pattern, $lastversion, $action) = split ' ', $line;
+	($site, $dir, $pattern, $lastversion, $action) = split ' ', $line, 5;
 
 	if (! defined $lastversion or $site =~ /\(.*\)/ or $dir =~ /\(.*\)/) {
 	    warn "$progname warning: there appears to be a version 2 format line in\n  the version 1 watchfile $watchfile;\n  Have you forgotten a 'version=2' line at the start, perhaps?\n  Skipping the line: $line\n";
@@ -651,24 +667,39 @@ sub process_watchline ($$$$$$)
 		elsif ($opt =~ /^dversionmangle\s*=\s*(.+)/) {
 		    @{$options{'dversionmangle'}} = split /;/, $1;
 		}
+		elsif ($opt =~ /^versionmangle\s*=\s*(.+)/) {
+		    @{$options{'uversionmangle'}} = split /;/, $1;
+		    @{$options{'dversionmangle'}} = split /;/, $1;
+		}
+		elsif ($opt =~ /^filenamemangle\s*=\s*(.+)/) {
+		    @{$options{'filenamemangle'}} = split /;/, $1;
+		}
+		elsif ($opt =~ /^downloadurlmangle\s*=\s*(.+)/) {
+		    @{$options{'downloadurlmangle'}} = split /;/, $1;
+		}
 		else {
 		    warn "$progname warning: unrecognised option $opt\n";
 		}
 	    }
 	}
 
-	($base, $filepattern, $lastversion, $action) = split ' ', $line;
+	($base, $filepattern, $lastversion, $action) = split ' ', $line, 4;
 
-	if ($base =~ m%/([^/]*\([^/]*\)[^/]*)$%) {
-	    # only three fields
-	    $action = $lastversion;
-	    $lastversion = $pattern;
+	if ($base =~ s%/([^/]*\([^/]*\)[^/]*)$%/%) {
+	    # Last component of $base has a pair of parentheses, so no
+	    # separate filepattern field; we remove the filepattern from the
+	    # end of $base and rescan the rest of the line
 	    $filepattern = $1;
-	    $base =~ s%/[^/]+$%/%;
+	    (undef, $lastversion, $action) = split ' ', $line, 3;
+	}
+
+	# Check validity of options
+	if ($base =~ /^ftp:/ and exists $options{'downloadurlmangle'}) {
+	    warn "$progname warning: downloadurlmangle option invalid for ftp sites,\n  ignoring in $watchfile:\n  $line\n";
 	}
 
 	# Handle sf.net addresses specially
-	$base =~ s%^http://sf.net/%http://qa.debian.org/watch/sf.php/%;
+	$base =~ s%^http://sf\.net/%http://qa.debian.org/watch/sf.php/%;
 	if ($base =~ m%^(\w+://[^/]+)%) {
 	    $site = $1;
 	} else {
@@ -690,7 +721,7 @@ sub process_watchline ($$$$$$)
 
     # Check all's OK
     if ($pattern !~ /\(.*\)/) {
-	warn "$progname warning: Filename pattern missing version delimeters ()\n  in $watchfile, skipping:\n  $line\n";
+	warn "$progname warning: Filename pattern missing version delimiters ()\n  in $watchfile, skipping:\n  $line\n";
 	return 1;
     }
 
@@ -838,10 +869,19 @@ EOF
     }
 
     my $newfile_base=basename($newfile);
+    if (exists $options{'filenamemangle'}) {
+        $newfile_base=$newfile;
+    }
+    foreach my $pat (@{$options{'filenamemangle'}}) {
+	eval "\$newfile_base =~ $pat;";
+    }
     # Remove HTTP header trash
-    if ($site =~ m%^http://%)
-    {
+    if ($site =~ m%^http://%) {
         $newfile_base =~ s/\?.*$//;
+	# just in case this leaves us with nothing
+	if ($newfile_base eq '') {
+	    $newfile_base = "$pkg-$newversion.download";
+	}
     }
     if (! $lastversion or $lastversion eq 'debian') {
 	$lastversion=$pkg_version;
@@ -868,6 +908,14 @@ EOF
 	else {
 	    $upstream_url = "$urlbase$newfile";
 	}
+
+	# mangle if necessary
+	$upstream_url =~ s/&amp;/&/g;
+	if (exists $options{'downloadurlmangle'}) {
+	    foreach my $pat (@{$options{'downloadurlmangle'}}) {
+		eval "\$upstream_url =~ $pat;";
+	    }
+	}
     }
     else {
 	# FTP site
@@ -961,6 +1009,8 @@ EOF
     print "-- Downloading updated package $newfile_base\n" if $verbose;
     # Download newer package
     if ($upstream_url =~ m%^http://%) {
+	# substitute HTML entities
+	# Is anything else than "&amp;" required?  I doubt it.
 	print STDERR "$progname debug: requesting URL $upstream_url\n" if $debug;
 	$request = HTTP::Request->new('GET', $upstream_url);
 	$response = $user_agent->request($request, "../$newfile_base");
@@ -987,6 +1037,15 @@ EOF
 	}
     }
 
+    if ($newfile_base =~ /\.(tar\.gz|tgz|tar\.bz2|tbz2?)$/) {
+	my $filetype = `file ../$newfile_base`;
+	$filetype =~ s%^\.\./\Q$newfile_base\E: %%;
+	unless ($filetype =~ /compressed data/) {
+	    warn "$progname warning: ../$newfile_base does not appear to be a compressed file;\nthe file command says: $filetype\nNot processing this file any further!\n";
+	    return 1;
+	}
+    }
+
     if ($newfile_base =~ /\.(tar\.gz|tgz)$/) {
 	if ($symlink eq 'symlink') {
 	    symlink $newfile_base, "../${pkg}_${newversion}.orig.tar.gz";
@@ -1065,6 +1124,11 @@ EOF
 	    $usefile = "../${pkg}_${newversion}.orig.tar.bz2";
 	}
 
+	# Any symlink requests are already handled by uscan
+	if ($action =~ /^uupdate(\s|$)/) {
+	    $action =~ s/^uupdate/uupdate --no-symlink/;
+	}
+
 	if ($watch_version > 1) {
 	    print "-- Executing user specified script\n     $action --upstream-version $newversion $newfile_base" if $verbose;
 	    if ($dehs) {
@@ -1334,15 +1398,15 @@ sub dehs_output ()
 	if (exists $dehs_tags{$tag}) {
 	    if (ref $dehs_tags{$tag} eq "ARRAY") {
 		foreach my $entry (@{$dehs_tags{$tag}}) {
-            $entry =~ s/</&lt;/g;
-            $entry =~ s/>/&gt;/g;
+		    $entry =~ s/</&lt;/g;
+		    $entry =~ s/>/&gt;/g;
 		    $entry =~ s/&/&amp;/g;
-            print "<$tag>$entry</$tag>\n";
+		    print "<$tag>$entry</$tag>\n";
 		}
 	    } else {
-            $dehs_tags{$tag} =~ s/</&lt;/g;
-            $dehs_tags{$tag} =~ s/>/&gt;/g;
-		    $dehs_tags{$tag} =~ s/&/&amp;/g;
+		$dehs_tags{$tag} =~ s/</&lt;/g;
+		$dehs_tags{$tag} =~ s/>/&gt;/g;
+		$dehs_tags{$tag} =~ s/&/&amp;/g;
 		print "<$tag>$dehs_tags{$tag}</$tag>\n";
 	    }
 	}
