@@ -34,7 +34,7 @@
 # number after the archive file name.  If both are given, the latter
 # takes precedence.
 #
-# The -u option requests that the new .orig.tar.gz archive be the
+# The -u option requests that the new .orig.tar.{gz|bz2} archive be the
 # pristine source, although this only makes sense when the original
 # archive itself is a tar.gz or tgz archive.
 #
@@ -57,10 +57,10 @@ Options are:
                       which command to be used to become root
                       for package-building
    --pristine, -u     Source is pristine upstream source and should be
-                      copied to <pkg>_<version>.orig.tar.gz; not valid
+                      copied to <pkg>_<version>.orig.tar.{gz|bz2}; not valid
                       for patches
-   --no-symlink       Copy new upstream .tar.gz archive to new location
-                      as <pkg>_<version>.orig.tar.gz instead of making a
+   --no-symlink       Copy new upstream archive to new location
+                      as <pkg>_<version>.orig.tar.{gz|bz2} instead of making a
                       symlink
    --no-conf, --noconf
                       Don't read devscripts config files;
@@ -167,11 +167,9 @@ TEMP=$(getopt -s bash -o v:p:r:ubs \
 	--long symlink,no-symlink,nosymlink \
 	--long no-conf,noconf \
 	--long help,version -n "$PROGNAME" -- "$@")
+if [ $? != 0 ] ; then exit 1 ; fi
 
-if [ $? != 0 ] ; then echo "Terminating..." >&2 ; exit 1 ; fi
-
-# Note the quotes around `$TEMP': they are essential!
-eval set -- "$TEMP"
+eval set -- $TEMP
 
 # Process Parameters
 while [ "$1" ]; do
@@ -202,7 +200,7 @@ while [ "$1" ]; do
 done
 
 if [ -n "$PATCH" ]; then
-    if [ $# ne 0 ]; then
+    if [ $# -ne 0 ]; then
         echo "$PROGNAME: additional archive name/version number is not allowed with --patch" >&2
 	echo "Run $PROGNAME --help for usage information" >&2
         exit 1
@@ -235,7 +233,7 @@ mustsetvar VERSION "`dpkg-parsechangelog | sed -n 's/^Version: //p'`" \
 # Get epoch and upstream version
 eval `echo "$VERSION" | perl -ne '/^(?:(\d+):)?(.*)/; print "SVERSION=$2\nEPOCH=$1\n";'`
 
-UVERSION=`expr "$SVERSION" : '\(.*\)-[0-9a-z.+]*$'` ||
+UVERSION=`expr "$SVERSION" : '\(.*\)-[0-9a-zA-Z.+]*$'` ||
 {
     echo "$PROGNAME: a native Debian package cannot take upstream updates" >&2
     exit 1
@@ -333,8 +331,16 @@ if [ "$PATCH" ]; then
 	echo "Aborting...." >&2
 	exit 1
     fi
-    if [ ! -r "../${PACKAGE}_$UVERSION.orig.tar.gz" ]; then
-	echo "$PROGNAME: can't find/read ${PACKAGE}_$UVERSION.orig.tar.gz" >&2
+
+    # Is the old version a .tar.gz or .tar.bz2 file?
+    if [ -r "../${PACKAGE}_$UVERSION.orig.tar.gz" ]; then
+	OLDARCHIVE="${PACKAGE}_$UVERSION.orig.tar.gz"
+	OLDARCHIVETYPE=gz
+    elif [ -r "../${PACKAGE}_$UVERSION.orig.tar.bz2" ]; then
+	OLDARCHIVE="${PACKAGE}_$UVERSION.orig.tar.bz2"
+	OLDARCHIVETYPE=bz2
+    else
+	echo "$PROGNAME: can't find/read ${PACKAGE}_$UVERSION.orig.tar.{gz|bz2}" >&2
 	echo "in the parent directory!" >&2
 	echo "Aborting...." >&2
 	exit 1
@@ -366,11 +372,23 @@ if [ "$PATCH" ]; then
 	exit 1
     }
     cd $TEMP_DIR
-    tar zxf ../${PACKAGE}_$UVERSION.orig.tar.gz || {
-	echo "$PROGNAME: can't untar ${PACKAGE}_$UVERSION.orig.tar.gz;" >&2
-	echo "aborting..." >&2
+    if [ "$OLDARCHIVETYPE" = gz ]; then
+	tar zxf ../$OLDARCHIVE || {
+	    echo "$PROGNAME: can't untar $OLDARCHIVE;" >&2
+	    echo "aborting..." >&2
+	    exit 1
+	}
+    elif [ "$OLDARCHIVETYPE" = bz2 ]; then
+	tar --bzip2 -xf ../$OLDARCHIVE || {
+	    echo "$PROGNAME: can't untar $OLDARCHIVE;" >&2
+	    echo "aborting..." >&2
+	    exit 1
+	}
+    else
+	echo "$PROGNAME: internal error: unknown OLDARCHIVETYPE: $OLDARCHIVETYPE" >&2
 	exit 1
-    }
+    fi
+
     if [ `ls | wc -l` -eq 1 ] && [ -d "`ls`" ]; then
 	mv "`ls`" ../${PACKAGE}-$UVERSION.orig
     else
@@ -443,19 +461,26 @@ else
     esac
 
     # Figure out the type of archive
-    X="${ARCHIVE##*/}"
-    case "$X" in
-	*.orig.tar.gz)  X="${X%.orig.tar.gz}";  UNPACK="tar zxf" ;;
-	*.tar.gz)  X="${X%.tar.gz}";  UNPACK="tar zxf" ;;
-	*.tar.bz2) X="${X%.tar.bz2}"; UNPACK="tar --bzip2 -xf" ;;
-	*.tar.Z)   X="${X%.tar.Z}";   UNPACK="tar zxf" ;;
-	*.tgz)     X="${X%.tgz}";     UNPACK="tar zxf" ;;
-	*.tar)     X="${X%.tar}";     UNPACK="tar xf"  ;;
-	*.zip)     X="${X%.zip}";     UNPACK="unzip"   ;;
-	*)
-	    echo "$PROGNAME: sorry: Unknown archive type" >&2
-	    exit 1
-    esac
+    X="${ARCHIVE%%/}"
+    X="${X##*/}"
+    if [ ! -d "$ARCHIVE_PATH" ]; then
+	case "$X" in
+	    *.orig.tar.gz)  X="${X%.orig.tar.gz}";  UNPACK="tar zxf";
+	                    TYPE=gz ;;
+	    *.orig.tar.bz2) X="${X%.orig.tar.bz2}"; UNPACK="tar --bzip -xf";
+	                    TYPE=bz2 ;;
+	    *.tar.gz)  X="${X%.tar.gz}";  UNPACK="tar zxf"; TYPE=gz ;;
+	    *.tar.bz2) X="${X%.tar.bz2}"; UNPACK="tar --bzip -xf"; TYPE=bz2 ;;
+	    *.tar.Z)   X="${X%.tar.Z}";   UNPACK="tar zxf"; TYPE="" ;;
+	    *.tgz)     X="${X%.tgz}";     UNPACK="tar zxf"; TYPE=gz ;;
+	    *.tar)     X="${X%.tar}";     UNPACK="tar xf";  TYPE="" ;;
+	    *.zip)     X="${X%.zip}";     UNPACK="unzip";   TYPE="" ;;
+	    *)
+		echo "$PROGNAME: sorry: Unknown archive type" >&2
+		exit 1
+	esac
+    fi
+
     if [ "$NEW_VERSION" = "" ]; then
 	# Figure out the new version
 	NEW_VERSION=`echo "$X" | perl -ne "/$MPATTERN/"' && print $1'`
@@ -494,42 +519,55 @@ else
 	exit 1
     fi
 
-    # Sanity check
+    # Sanity checks
     if [ -e "../${PACKAGE}_$SNEW_VERSION.orig.tar.gz" ] && \
-	[ `md5sum "${ARCHIVE_PATH}" | cut -d" " -f1` != \
-	  `md5sum "../${PACKAGE}_$SNEW_VERSION.orig.tar.gz" | cut -d" " -f1` ]
+	[ "$(md5sum "${ARCHIVE_PATH}" | cut -d" " -f1)" != \
+	  "$(md5sum "../${PACKAGE}_$SNEW_VERSION.orig.tar.gz" | cut -d" " -f1)" ]
     then
 	echo "$PROGNAME: a different ${PACKAGE}_$SNEW_VERSION.orig.tar.gz" >&2
 	echo "already exists in the parent dir;" >&2
-	echo " please check on the situation before trying $PROGNAME again." >&2
+	echo "please check on the situation before trying $PROGNAME again." >&2
+	exit 1
+    elif [ -e "../${PACKAGE}_$SNEW_VERSION.orig.tar.bz2" ] && \
+	[ "$(md5sum "${ARCHIVE_PATH}" | cut -d" " -f1)" != \
+	  "$(md5sum "../${PACKAGE}_$SNEW_VERSION.orig.tar.bz2" | cut -d" " -f1)" ]
+    then
+	echo "$PROGNAME: a different ${PACKAGE}_$SNEW_VERSION.orig.tar.bz2" >&2
+	echo "already exists in the parent dir;" >&2
+	echo "please check on the situation before trying $PROGNAME again." >&2
 	exit 1
     fi
 
-    if [ $UUPDATE_PRISTINE = yes -a \
-	! -e "../${PACKAGE}_$SNEW_VERSION.orig.tar.gz" ]; then
+    if [ $UUPDATE_PRISTINE = yes -a -n "$TYPE" -a \
+	! -e "../${PACKAGE}_$SNEW_VERSION.orig.tar.gz" -a \
+	! -e "../${PACKAGE}_$SNEW_VERSION.orig.tar.bz2" ]; then
 	if [ "$UUPDATE_SYMLINK_ORIG" = yes ]; then
-	    echo "Symlinking to pristine source from ${PACKAGE}_$SNEW_VERSION.orig.tar.gz..."
+	    echo "Symlinking to pristine source from ${PACKAGE}_$SNEW_VERSION.orig.tar.$TYPE..."
+	    case $ARCHIVE_PATH in
+		/*)   LINKARCHIVE="$ARCHIVE" ;;
+		../*) LINKARCHIVE="${ARCHIVE#../}" ;;
+	    esac
 	else
-	    echo "Copying pristine source to ${PACKAGE}_$SNEW_VERSION.orig.tar.gz..."
+	    echo "Copying pristine source to ${PACKAGE}_$SNEW_VERSION.orig.tar.$TYPE..."
 	fi
-	case $ARCHIVE_PATH in
-	    /*.tar.gz|/*.tgz)
+
+	case "$TYPE" in
+	    gz)
 		if [ "$UUPDATE_SYMLINK_ORIG" = yes ]; then
-		    ln -s "$ARCHIVE" "../${PACKAGE}_$SNEW_VERSION.orig.tar.gz"
+		    ln -s "$LINKARCHIVE" "../${PACKAGE}_$SNEW_VERSION.orig.tar.gz"
 		else
-		    cp "$ARCHIVE" "../${PACKAGE}_$SNEW_VERSION.orig.tar.gz"
+		    cp "$ARCHIVE_PATH" "../${PACKAGE}_$SNEW_VERSION.orig.tar.gz"
 		fi
 		;;
-	    ../*.tar.gz|../*.tgz)
+	    bz2)
 		if [ "$UUPDATE_SYMLINK_ORIG" = yes ]; then
-		    ln -s "${ARCHIVE#../}" "../${PACKAGE}_$SNEW_VERSION.orig.tar.gz"
+		    ln -s "$LINKARCHIVE" "../${PACKAGE}_$SNEW_VERSION.orig.tar.bz2"
 		else
-		    cp "${ARCHIVE#../}" "../${PACKAGE}_$SNEW_VERSION.orig.tar.gz"
+		    cp "$ARCHIVE_PATH" "../${PACKAGE}_$SNEW_VERSION.orig.tar.bz2"
 		fi
 		;;
-		# no other ARCHIVE_PATH possibilities: either absolute or ../
 	    *)
-		echo "$PROGNAME: can't preserve pristine sources from non .tar.gz upstream archive!" >&2
+		echo "$PROGNAME: can't preserve pristine sources from non .tar.gz/.tar.bz2 upstream archive!" >&2
 		echo "Continuing anyway..." >&2
 		;;
 	esac
@@ -542,12 +580,21 @@ else
 	exit 1
     }
     cd $TEMP_DIR
-    echo "-- Untarring the new sourcecode archive $ARCHIVE"
-    $UNPACK "$ARCHIVE_PATH" || {
-	echo "$PROGNAME: can't unpack: $UNPACK $ARCHIVE_PATH failed;" >&2
-	echo "aborting..." >&2
-	exit 1
-    }
+    if [ ! -d "$ARCHIVE_PATH" ]; then
+	echo "-- Untarring the new sourcecode archive $ARCHIVE"
+	$UNPACK "$ARCHIVE_PATH" || {
+	    echo "$PROGNAME: can't unpack: $UNPACK $ARCHIVE_PATH failed;" >&2
+	    echo "aborting..." >&2
+	    exit 1
+	}
+    else
+	tar -C "$ARCHIVE_PATH/../" -c $X | tar x || {
+	    echo "$PROGNAME: tar -C \"$ARCHIVE_PATH/../\" -c $X | tar x failed;" >&2
+	    echo "aborting..." >&2
+	    exit 1
+	}
+    fi
+
     cd ..
     if [ `ls $TEMP_DIR | wc -l` -eq 1 ]; then
 	# The files are stored in the archive under a top directory, we presume
@@ -560,11 +607,29 @@ else
     rm -rf $TEMP_DIR
     cp -a $PACKAGE-$SNEW_VERSION $PACKAGE-$SNEW_VERSION.orig
     cd $PACKAGE-$SNEW_VERSION
-    if [ -r "../${PACKAGE}_$SVERSION.diff.gz" ];
-    then
+
+    if [ -r "../${PACKAGE}_$SVERSION.diff.gz" ]; then
+	DIFF="../${PACKAGE}_$SVERSION.diff.gz"
+	DIFFTYPE=diff
+	DIFFCAT=zcat
+    elif [ -r "../${PACKAGE}_$SVERSION.diff.bz2" ]; then
+	DIFF="../${PACKAGE}_$SVERSION.diff.bz2"
+	DIFFTYPE=diff
+	DIFFCAT=bzcat
+    elif [ -r "../${PACKAGE}_$SVERSION.debian.tar.gz" ]; then
+	DIFF="../${PACKAGE}_$SVERSION.debian.tar.gz"
+	DIFFTYPE=tar
+	DIFFUNPACK="tar zxf"
+    elif [ -r "../${PACKAGE}_$SVERSION.debian.tar.bz2" ]; then
+	DIFF="../${PACKAGE}_$SVERSION.debian.tar.bz2"
+	DIFFTYPE=tar
+	DIFFUNPACK="tar --bzip2 -xf"
+    fi
+
+    if [ "$DIFFTYPE" = diff ]; then
 	# Check that any files added in diff do not now exist in
 	# upstream version
-	FILES=$(zcat ../${PACKAGE}_$SVERSION.diff.gz |
+	FILES=$($DIFFCAT $DIFF |
 	        perl -nwe 'BEGIN { $status=""; }
 	                   chomp;
 	                   if (/^--- /) { $status = "-$."; }
@@ -579,42 +644,40 @@ else
 	                       }
 	                   }')
 
-	# Note that debian/changelog is always in FILES, so FILES is non-null
-	# Check anyway!
-	if [ -z "$FILES" ]; then
-	    echo "$PROGNAME: FILES is empty - why?  Please file a bug report!" >&2
-	    echo "Aborting..." >&2
-	    exit 1
-	fi
-
-	for file in $FILES; do
-	    if [ -e "$file" ]; then
-		echo "$PROGNAME warning: file $file was added in old diff, but is now in the upstream source." >&2
-		echo "Please check that the diff is applied correctly." >&2
-		echo "(This program will use the pristine upstream version and save the old .diff.gz" >&2
-		echo "version as $file.debdiff .)" >&2
-
-		if [ -e "$file.upstream" -o -e "$file.debdiff" ]; then
-		    FILEEXISTERR=1
+	# Note that debian/changelog is usually in FILES, so FILES is
+	# usually non-null; however, if the upstream ships its own debian/
+	# directory, this may not be true, so must check for empty $FILES.
+	# Check anyway, even though it's not strictly necessary in bash.
+	if [ -n "$FILES" ]; then
+	    for file in $FILES; do
+		if [ -e "$file" ]; then
+		    echo "$PROGNAME warning: file $file was added in old diff, but is now in the upstream source." >&2
+		    echo "Please check that the diff is applied correctly." >&2
+		    echo "(This program will use the pristine upstream version and save the old .diff.gz" >&2
+		    echo "version as $file.debdiff .)" >&2
+		    
+		    if [ -e "$file.upstream" -o -e "$file.debdiff" ]; then
+			FILEEXISTERR=1
+		    fi
 		fi
-	    fi
-	done
+	    done
 
-	if [ -n "$FILEEXISTERR" ]; then
-	    echo "$PROGNAME: please apply the diff by hand and take care with this." >&2
-	    exit 1
+	    if [ -n "$FILEEXISTERR" ]; then
+		echo "$PROGNAME: please apply the diff by hand and take care with this." >&2
+		exit 1
+	    fi
+
+	    # Shift any files that are in the upstream tarball that are also in
+	    # the old diff out of the way so the diff is more likely to apply
+	    # cleanly, and remember the fact that we moved it
+	    for file in $FILES; do
+		if [ -e "$file" ]; then
+		    mv $file $file.upstream
+		    MOVEDFILES=("${MOVEDFILES[@]}" "$file")
+		fi
+	    done
 	fi
 
-	# Shift any files that are in the upstream tarball that are also in
-	# the old diff out of the way so the diff is more likely to apply
-	# cleanly, and remember the fact that we moved it
-	for file in $FILES; do
-	    if [ -e "$file" ]; then
-		mv $file $file.upstream
-		MOVEDFILES=("${MOVEDFILES[@]}" "$file")
-	    fi
-	done
-
 	if zcat ../${PACKAGE}_$SVERSION.diff.gz | patch -sNp1 ; then
 	    echo "Success!  The diffs from version $VERSION worked fine."
 	else
@@ -633,6 +696,24 @@ else
 	    fi
 	done
 
+    elif [ "$DIFFTYPE" = tar ]; then
+	if [ -d debian ]; then
+	    echo "$PROGNAME warning: using a debian.tar.{gz|bz2} file in old Debian source," >&@
+	    echo "but upstream also contains a debian/ directory!" >&2
+	    if [ -e "debian.upstream" ]; then
+		echo "Please apply the diff by hand and take care with this." >&2
+		exit 1
+	    fi
+	    echo "This program will move the upstream directory out of the way" >&2
+	    echo "to debian.upstream/ and use the Debian version" >&2
+	    mv debian debian.upstream
+	fi
+	if $DIFFUNPACK $DIFF; then
+	    echo "Unpacking the debian/ directory from version $VERSION worked fine."
+	else 
+	    echo "$PROGNAME: failed to unpack the debian/ directory from version $VERSION!" >&2
+	    exit 1
+	fi
     else
 	echo "$PROGNAME: could not find diffs from version $VERSION to apply!" >&2
 	exit 1
